@@ -11,6 +11,7 @@
 #include "nsRect.h"
 #include "nsRectAbsolute.h"
 #include "gfxRect.h"
+#include "mozilla/gfx/Point.h"
 #include "mozilla/gfx/Rect.h"
 #include "mozilla/gfx/RectAbsolute.h"
 #include "mozilla/WritingModes.h"
@@ -18,11 +19,14 @@
 #  include <windows.h>
 #endif
 
+using mozilla::CSSCoord;
 using mozilla::CSSIntCoord;
 using mozilla::CSSIntSize;
 using mozilla::ScreenIntCoord;
+using mozilla::gfx::IntPoint;
 using mozilla::gfx::IntRect;
 using mozilla::gfx::IntRectAbsolute;
+using mozilla::gfx::Rect;
 
 static_assert(std::is_constructible_v<CSSIntSize, CSSIntCoord, CSSIntCoord>);
 static_assert(
@@ -30,6 +34,11 @@ static_assert(
 static_assert(std::is_constructible_v<CSSIntSize, int, int>);
 static_assert(!std::is_constructible_v<CSSIntSize, float, float>);
 
+static_assert(std::is_same_v<CSSIntCoord, decltype(CSSIntCoord() * 42)>);
+static_assert(std::is_same_v<CSSCoord, decltype(CSSCoord() * 42)>);
+static_assert(std::is_same_v<CSSCoord, decltype(CSSIntCoord() * 42.f)>);
+static_assert(std::is_same_v<CSSCoord, decltype(CSSCoord() * 42.f)>);
+
 template <class RectType>
 static bool TestConstructors() {
   // Create a rectangle
@@ -636,10 +645,14 @@ TEST(Gfx, gfxRect)
 }
 
 TEST(Gfx, nsRectAbsolute)
-{ TestUnionEmptyRects<nsRectAbsolute>(); }
+{
+  TestUnionEmptyRects<nsRectAbsolute>();
+}
 
 TEST(Gfx, IntRectAbsolute)
-{ TestUnionEmptyRects<IntRectAbsolute>(); }
+{
+  TestUnionEmptyRects<IntRectAbsolute>();
+}
 
 static void TestMoveInsideAndClamp(IntRect aSrc, IntRect aTarget,
                                    IntRect aExpected) {
@@ -674,3 +687,51 @@ TEST(Gfx, MoveInsideAndClamp)
   TestMoveInsideAndClamp(IntRect(0, 0, 0, 0), IntRect(10, -10, 10, 10),
                          IntRect(10, 0, 0, 0));
 }
+
+TEST(Gfx, ClampPoint)
+{
+  /* Various bounds */
+  IntRect Square(0, 0, 10, 10);
+  EXPECT_EQ(Square.ClampPoint(IntPoint(-5, -5)), IntPoint(0, 0));
+  EXPECT_EQ(Square.ClampPoint(IntPoint(-5, 5)), IntPoint(0, 5));
+  EXPECT_EQ(Square.ClampPoint(IntPoint(-5, 15)), IntPoint(0, 10));
+
+  EXPECT_EQ(Square.ClampPoint(IntPoint(5, -5)), IntPoint(5, 0));
+  EXPECT_EQ(Square.ClampPoint(IntPoint(5, 5)), IntPoint(5, 5));
+  EXPECT_EQ(Square.ClampPoint(IntPoint(5, 15)), IntPoint(5, 10));
+
+  EXPECT_EQ(Square.ClampPoint(IntPoint(15, -5)), IntPoint(10, 0));
+  EXPECT_EQ(Square.ClampPoint(IntPoint(15, 5)), IntPoint(10, 5));
+  EXPECT_EQ(Square.ClampPoint(IntPoint(15, 15)), IntPoint(10, 10));
+
+  /* Special case of empty rect */
+  IntRect Empty(0, 0, -1, -1);
+  EXPECT_TRUE(Empty.IsEmpty());
+  EXPECT_EQ(Empty.ClampPoint(IntPoint(-1, -1)), IntPoint(0, 0));
+  EXPECT_EQ(Empty.ClampPoint(IntPoint(-1, 1)), IntPoint(0, 0));
+  EXPECT_EQ(Empty.ClampPoint(IntPoint(1, -1)), IntPoint(0, 0));
+  EXPECT_EQ(Empty.ClampPoint(IntPoint(1, 1)), IntPoint(0, 0));
+}
+
+TEST(Gfx, SafeMoveBy)
+{
+  IntRect intRect(0, 0, 10, 10);
+  intRect.SafeMoveByX(10);
+  intRect.SafeMoveByY(10);
+  EXPECT_EQ(intRect, IntRect(10, 10, 10, 10));
+
+  intRect = IntRect(0, 0, 10, 10);
+  intRect.SafeMoveByX(-10);
+  intRect.SafeMoveByY(-10);
+  EXPECT_EQ(intRect, IntRect(-10, -10, 10, 10));
+
+  Rect rect(0, 0, 10, 10);
+  rect.SafeMoveByX(10);
+  rect.SafeMoveByY(10);
+  EXPECT_EQ(rect, Rect(10, 10, 10, 10));
+
+  rect = Rect(0, 0, 10, 10);
+  rect.SafeMoveByX(-10);
+  rect.SafeMoveByY(-10);
+  EXPECT_EQ(rect, Rect(-10, -10, 10, 10));
+}
