@@ -11,7 +11,7 @@
 #include "mozilla/D3DMessageUtils.h"
 #include "mozilla/StaticPrefs_gfx.h"
 #include "mozilla/StaticPrefs_layers.h"
-#include "mozilla/Telemetry.h"
+#include "mozilla/glean/GfxMetrics.h"
 #include "mozilla/gfx/GPUParent.h"
 #include "mozilla/gfx/GPUProcessManager.h"
 #include "mozilla/gfx/GraphicsMessages.h"
@@ -26,8 +26,8 @@
 
 // -
 
-#include "mozilla/gfx/AllOfDcomp.h"
 #include <d3d11.h>
+#include <dcomp.h>
 #include <ddraw.h>
 #include <dxgi.h>
 
@@ -52,11 +52,6 @@ decltype(DCompositionCreateDevice3)* sDcompCreateDevice3Fn = nullptr;
 decltype(DCompositionCreateSurfaceHandle)* sDcompCreateSurfaceHandleFn =
     nullptr;
 
-// We don't have access to the DirectDrawCreateEx type in gfxWindowsPlatform.h,
-// since it doesn't include ddraw.h, so we use a static here. It should only
-// be used within InitializeDirectDrawConfig.
-decltype(DirectDrawCreateEx)* sDirectDrawCreateExFn = nullptr;
-
 /* static */
 void DeviceManagerDx::Init() { sInstance = new DeviceManagerDx(); }
 
@@ -206,6 +201,129 @@ bool DeviceManagerDx::GetOutputFromMonitor(HMONITOR monitor,
   return false;
 }
 
+void DeviceManagerDx::PostUpdateMonitorInfo() {
+  MOZ_ASSERT(XRE_IsGPUProcess());
+  MOZ_ASSERT(NS_IsMainThread());
+
+  MutexAutoLock lock(mDeviceLock);
+  // Reduce frequency of UpdateMonitorInfo() call.
+  if (mUpdateMonitorInfoRunnable) {
+    return;
+  }
+
+  auto* holder = CompositorThreadHolder::GetSingleton();
+  if (!holder) {
+    return;
+  }
+
+  mUpdateMonitorInfoRunnable = NS_NewRunnableFunction(
+      "DeviceManagerDx::PostUpdateMonitorInfo::Runnable", []() -> void {
+        auto* dm = gfx::DeviceManagerDx::Get();
+        if (dm) {
+          dm->UpdateMonitorInfo();
+        }
+      });
+
+  const uint32_t kDelayMS = 100;
+  RefPtr<Runnable> runnable = mUpdateMonitorInfoRunnable;
+  holder->GetCompositorThread()->DelayedDispatch(runnable.forget(), kDelayMS);
+}
+
+static bool ColorSpaceIsHDR(const DXGI_OUTPUT_DESC1& aDesc) {
+  // Set isHDR to true if the output has a BT2020 colorspace with EOTF2084
+  // gamma curve, this indicates the system is sending an HDR format to
+  // this monitor.  The colorspace returned by DXGI is very vague - we only
+  // see DXGI_COLOR_SPACE_RGB_FULL_G2084_NONE_P2020 for HDR and
+  // DXGI_COLOR_SPACE_RGB_FULL_G22_NONE_P709 for SDR modes, even if the
+  // monitor is using something like YCbCr444 according to Settings
+  // (System -> Display Settings -> Advanced Display).  To get more specific
+  // info we would need to query the DISPLAYCONFIG values in WinGDI.
+  //
+  // Note that we don't check bit depth here, since as of Windows 11 22H2,
+  // HDR is supported with 8bpc for lower bandwidth, where DWM converts to
+  // dithered RGB8 rather than RGB10, which doesn't really matter here.
+  //
+  // Since RefreshScreens(), the caller of this function, is triggered
+  // by WM_DISPLAYCHANGE, this will pick up changes to the monitors in
+  // all the important cases (resolution/color changes by the user).
+  //
+  // Further reading:
+  // https://learn.microsoft.com/en-us/windows/win32/direct3darticles/high-dynamic-range
+  // https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-displayconfig_sdr_white_level
+  bool isHDR = (aDesc.ColorSpace == DXGI_COLOR_SPACE_RGB_FULL_G2084_NONE_P2020);
+
+  return isHDR;
+}
+
+void DeviceManagerDx::UpdateMonitorInfo() {
+  bool systemHdrEnabled = false;
+  std::set<HMONITOR> hdrMonitors;
+
+  for (const auto desc : EnumerateOutputs()) {
+    if (ColorSpaceIsHDR(desc)) {
+      systemHdrEnabled = true;
+      hdrMonitors.emplace(desc.Monitor);
+    }
+  }
+
+  {
+    MutexAutoLock lock(mDeviceLock);
+    mSystemHdrEnabled = Some(systemHdrEnabled);
+    mHdrMonitors.swap(hdrMonitors);
+    mUpdateMonitorInfoRunnable = nullptr;
+  }
+}
+
+bool DeviceManagerDx::SystemHDREnabled() {
+  {
+    MutexAutoLock lock(mDeviceLock);
+    if (mSystemHdrEnabled.isSome()) {
+      return mSystemHdrEnabled.ref();
+    }
+  }
+
+  UpdateMonitorInfo();
+
+  MutexAutoLock lock(mDeviceLock);
+  return mSystemHdrEnabled.ref();
+}
+
+bool DeviceManagerDx::WindowHDREnabled(HWND aWindow) {
+  MOZ_ASSERT(aWindow);
+
+  HMONITOR monitor = ::MonitorFromWindow(aWindow, MONITOR_DEFAULTTONEAREST);
+  return MonitorHDREnabled(monitor);
+}
+
+bool DeviceManagerDx::MonitorHDREnabled(HMONITOR aMonitor) {
+  if (!aMonitor) {
+    return false;
+  }
+
+  bool needInit = false;
+
+  {
+    MutexAutoLock lock(mDeviceLock);
+    if (mSystemHdrEnabled.isNothing()) {
+      needInit = true;
+    }
+  }
+
+  if (needInit) {
+    UpdateMonitorInfo();
+  }
+
+  MutexAutoLock lock(mDeviceLock);
+  MOZ_ASSERT(mSystemHdrEnabled.isSome());
+
+  auto it = mHdrMonitors.find(aMonitor);
+  if (it == mHdrMonitors.end()) {
+    return false;
+  }
+
+  return true;
+}
+
 void DeviceManagerDx::CheckHardwareStretchingSupport(HwStretchingSupport& aRv) {
   RefPtr<IDXGIAdapter> adapter = GetDXGIAdapter();
 
@@ -386,18 +504,6 @@ bool DeviceManagerDx::CreateCanvasDeviceLocked() {
     return false;
   }
 
-  if (StaticPrefs::
-          gfx_direct2d_target_independent_rasterization_disabled_AtStartup()) {
-    int creationFlags = 0x2;  // disable target independent rasterization
-    const GUID D2D_INTERNAL_DEVICE_CREATION_OPTIONS = {
-        0xfb3a8e1a,
-        0x2e3c,
-        0x4de1,
-        {0x84, 0x42, 0x40, 0x43, 0xe0, 0xb0, 0x94, 0x95}};
-    mCanvasDevice->SetPrivateData(D2D_INTERNAL_DEVICE_CREATION_OPTIONS,
-                                  sizeof(creationFlags), &creationFlags);
-  }
-
   if (FAILED(hr) || !mCanvasDevice) {
     NS_WARNING("Failed to acquire a D3D11 device for Canvas");
     return false;
@@ -515,7 +621,7 @@ void DeviceManagerDx::CreateContentDevicesLocked() {
 
   // We should have been assigned a DeviceStatus from the parent process,
   // GPU process, or the same process if using in-process compositing.
-  MOZ_ASSERT(mDeviceStatus);
+  MOZ_RELEASE_ASSERT(mDeviceStatus);
 
   if (CreateContentDevice() == FeatureStatus::CrashedInHandler) {
     DisableD3D11AfterCrash();
@@ -528,9 +634,11 @@ already_AddRefed<IDXGIAdapter1> DeviceManagerDx::GetDXGIAdapter() {
 }
 
 IDXGIAdapter1* DeviceManagerDx::GetDXGIAdapterLocked() {
-  if (mAdapter) {
+  if (mAdapter && mFactory && mFactory->IsCurrent()) {
     return mAdapter;
   }
+  mAdapter = nullptr;
+  mFactory = nullptr;
 
   nsModuleHandle dxgiModule(LoadLibrarySystem32(L"dxgi.dll"));
   decltype(CreateDXGIFactory1)* createDXGIFactory1 =
@@ -545,50 +653,32 @@ IDXGIAdapter1* DeviceManagerDx::GetDXGIAdapterLocked() {
 
   // Try to use a DXGI 1.1 adapter in order to share resources
   // across processes.
-  RefPtr<IDXGIFactory1> factory1;
   if (StaticPrefs::gfx_direct3d11_enable_debug_layer_AtStartup()) {
-    RefPtr<IDXGIFactory2> factory2;
     if (fCreateDXGIFactory2) {
       auto hr = fCreateDXGIFactory2(DXGI_CREATE_FACTORY_DEBUG,
                                     __uuidof(IDXGIFactory2),
-                                    getter_AddRefs(factory2));
+                                    getter_AddRefs(mFactory));
       MOZ_ALWAYS_TRUE(!FAILED(hr));
     } else {
       NS_WARNING(
           "fCreateDXGIFactory2 not loaded, cannot create debug IDXGIFactory2.");
     }
-    factory1 = factory2;
   }
-  if (!factory1) {
+  if (!mFactory) {
     HRESULT hr =
-        createDXGIFactory1(__uuidof(IDXGIFactory1), getter_AddRefs(factory1));
-    if (FAILED(hr) || !factory1) {
+        createDXGIFactory1(__uuidof(IDXGIFactory1), getter_AddRefs(mFactory));
+    if (FAILED(hr) || !mFactory) {
       // This seems to happen with some people running the iZ3D driver.
       // They won't get acceleration.
       return nullptr;
     }
   }
 
-  if (!mDeviceStatus) {
-    // If we haven't created a device yet, and have no existing device status,
-    // then this must be the compositor device. Pick the first adapter we can.
-    if (FAILED(factory1->EnumAdapters1(0, getter_AddRefs(mAdapter)))) {
-      return nullptr;
-    }
-  } else {
-    // In the UI and GPU process, we clear mDeviceStatus on device reset, so we
-    // should never reach here. Furthermore, the UI process does not create
-    // devices when using a GPU process.
-    //
-    // So, this should only ever get called on the content process or RDD
-    // process
-    MOZ_ASSERT(XRE_IsContentProcess() || XRE_IsRDDProcess());
-
-    // In the child process, we search for the adapter that matches the parent
-    // process. The first adapter can be mismatched on dual-GPU systems.
+  if (mDeviceStatus) {
+    // Match the adapter to our mDeviceStatus, if possible.
     for (UINT index = 0;; index++) {
       RefPtr<IDXGIAdapter1> adapter;
-      if (FAILED(factory1->EnumAdapters1(index, getter_AddRefs(adapter)))) {
+      if (FAILED(mFactory->EnumAdapters1(index, getter_AddRefs(adapter)))) {
         break;
       }
 
@@ -607,7 +697,9 @@ IDXGIAdapter1* DeviceManagerDx::GetDXGIAdapterLocked() {
   }
 
   if (!mAdapter) {
-    return nullptr;
+    mDeviceStatus.reset();
+    // Pick the first adapter available.
+    mFactory->EnumAdapters1(0, getter_AddRefs(mAdapter));
   }
 
   // We leak this module everywhere, we might as well do so here as well.
@@ -740,6 +832,13 @@ void DeviceManagerDx::CreateCompositorDevice(FeatureState& d3d11) {
     D3D11Checks::WarnOnAdapterMismatch(device);
   }
 
+  RefPtr<ID3D10Multithread> multi;
+  HRESULT hr = device->QueryInterface(__uuidof(ID3D10Multithread),
+                                      getter_AddRefs(multi));
+  if (SUCCEEDED(hr) && multi) {
+    multi->SetMultithreadProtected(TRUE);
+  }
+
   uint32_t featureLevel = device->GetFeatureLevel();
   auto formatOptions = D3D11Checks::FormatOptions(device);
   mCompositorDevice = device;
@@ -789,7 +888,7 @@ bool DeviceManagerDx::CreateDevice(IDXGIAdapter* aAdapter,
       // problematic.
       D3D11_MESSAGE_ID blockIDs[] = {
           D3D11_MESSAGE_ID_DEVICE_DRAW_CONSTANT_BUFFER_TOO_SMALL};
-      filter.DenyList.NumIDs = MOZ_ARRAY_LENGTH(blockIDs);
+      filter.DenyList.NumIDs = std::size(blockIDs);
       filter.DenyList.pIDList = blockIDs;
       infoQueue->PushStorageFilter(&filter);
 
@@ -832,6 +931,13 @@ void DeviceManagerDx::CreateWARPCompositorDevice() {
 
   bool textureSharingWorks = D3D11Checks::DoesTextureSharingWork(device);
 
+  RefPtr<ID3D10Multithread> multi;
+  hr = device->QueryInterface(__uuidof(ID3D10Multithread),
+                              getter_AddRefs(multi));
+  if (SUCCEEDED(hr) && multi) {
+    multi->SetMultithreadProtected(TRUE);
+  }
+
   DXGI_ADAPTER_DESC desc;
   D3D11Checks::GetDxgiDesc(device, &desc);
 
@@ -851,7 +957,7 @@ void DeviceManagerDx::CreateWARPCompositorDevice() {
 
 FeatureStatus DeviceManagerDx::CreateContentDevice() {
   RefPtr<IDXGIAdapter1> adapter;
-  if (!mDeviceStatus->isWARP()) {
+  if (!IsWARPLocked()) {
     adapter = GetDXGIAdapterLocked();
     if (!adapter) {
       gfxCriticalNote << "Could not get a DXGI adapter";
@@ -864,7 +970,7 @@ FeatureStatus DeviceManagerDx::CreateContentDevice() {
 
   UINT flags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;
   D3D_DRIVER_TYPE type =
-      mDeviceStatus->isWARP() ? D3D_DRIVER_TYPE_WARP : D3D_DRIVER_TYPE_UNKNOWN;
+      IsWARPLocked() ? D3D_DRIVER_TYPE_WARP : D3D_DRIVER_TYPE_UNKNOWN;
   if (!CreateDevice(adapter, type, flags, hr, device)) {
     gfxCriticalNote
         << "Recovered from crash while creating a D3D11 content device";
@@ -908,7 +1014,7 @@ FeatureStatus DeviceManagerDx::CreateContentDevice() {
 }
 
 RefPtr<ID3D11Device> DeviceManagerDx::CreateDecoderDevice(
-    bool aHardwareWebRender) {
+    DeviceFlagSet aFlags) {
   MutexAutoLock lock(mDeviceLock);
 
   if (!mDeviceStatus) {
@@ -926,15 +1032,16 @@ RefPtr<ID3D11Device> DeviceManagerDx::CreateDecoderDevice(
 
   if (reuseDevice) {
     // Use mCompositorDevice for decoder device only for hardware WebRender.
-    if (aHardwareWebRender && mCompositorDevice &&
-        mCompositorDeviceSupportsVideo && !mDecoderDevice) {
+    if (aFlags.contains(DeviceFlag::isHardwareWebRenderInUse) &&
+        mCompositorDevice && mCompositorDeviceSupportsVideo &&
+        !mDecoderDevice) {
       mDecoderDevice = mCompositorDevice;
 
       RefPtr<ID3D10Multithread> multi;
       mDecoderDevice->QueryInterface(__uuidof(ID3D10Multithread),
                                      getter_AddRefs(multi));
       if (multi) {
-        multi->SetMultithreadProtected(TRUE);
+        MOZ_ASSERT(multi->GetMultithreadProtected());
       }
     }
 
@@ -1002,7 +1109,7 @@ RefPtr<ID3D11Device> DeviceManagerDx::CreateMediaEngineDevice() {
   if (FAILED(hr) || !device || !D3D11Checks::DoesDeviceWork()) {
     return nullptr;
   }
-  Unused << SetDebugName(device.get(), "MFMediaEngineDevice");
+  (void)SetDebugName(device.get(), "MFMediaEngineDevice");
 
   RefPtr<ID3D10Multithread> multi;
   device->QueryInterface(__uuidof(ID3D10Multithread), getter_AddRefs(multi));
@@ -1073,6 +1180,10 @@ bool DeviceManagerDx::ContentAdapterIsParentAdapter(ID3D11Device* device) {
     return false;
   }
 
+  if (!mDeviceStatus) {
+    return false;
+  }
+
   const DxgiAdapterDesc& preferred = mDeviceStatus->adapter();
 
   if (desc.VendorId != preferred.VendorId ||
@@ -1165,8 +1276,8 @@ bool DeviceManagerDx::GetAnyDeviceRemovedReason(DeviceResetReason* aOutReason) {
 }
 
 void DeviceManagerDx::ForceDeviceReset(ForcedDeviceResetReason aReason) {
-  Telemetry::Accumulate(Telemetry::FORCED_DEVICE_RESET_REASON,
-                        uint32_t(aReason));
+  glean::gfx::forced_device_reset_reason.AccumulateSingleSample(
+      uint32_t(aReason));
   {
     MutexAutoLock lock(mDeviceLock);
     if (!mDeviceResetReason) {
@@ -1184,6 +1295,10 @@ void DeviceManagerDx::DisableD3D11AfterCrash() {
 }
 
 RefPtr<ID3D11Device> DeviceManagerDx::GetCompositorDevice() {
+  /// ID3D11Device is thread-safe. We need the lock to read the
+  /// mDeviceLockPointer, but manipulating the pointee outside of the lock is
+  /// safe. See
+  /// https://learn.microsoft.com/en-us/windows/win32/direct3d11/overviews-direct3d-11-render-multi-thread-intro
   MutexAutoLock lock(mDeviceLock);
   return mCompositorDevice;
 }
@@ -1217,8 +1332,9 @@ RefPtr<ID3D11Device> DeviceManagerDx::GetImageDevice() {
   if (FAILED(hr) || !multi) {
     gfxWarning() << "Multithread safety interface not supported. " << hr;
     return nullptr;
+  } else {
+    MOZ_ASSERT(multi->GetMultithreadProtected());
   }
-  multi->SetMultithreadProtected(TRUE);
 
   mImageDevice = device;
 
@@ -1267,6 +1383,10 @@ bool DeviceManagerDx::CanInitializeKeyedMutexTextures() {
 
 bool DeviceManagerDx::IsWARP() {
   MutexAutoLock lock(mDeviceLock);
+  return IsWARPLocked();
+}
+
+bool DeviceManagerDx::IsWARPLocked() {
   if (!mDeviceStatus) {
     return false;
   }
@@ -1305,58 +1425,6 @@ bool DeviceManagerDx::CanUseDComp() {
   return !!mDirectCompositionDevice;
 }
 
-void DeviceManagerDx::InitializeDirectDraw() {
-  MOZ_ASSERT(layers::CompositorThreadHolder::IsInCompositorThread());
-
-  if (mDirectDraw) {
-    // Already initialized.
-    return;
-  }
-
-  FeatureState& ddraw = gfxConfig::GetFeature(Feature::DIRECT_DRAW);
-  if (!ddraw.IsEnabled()) {
-    return;
-  }
-
-  // Check if DirectDraw is available on this system.
-  mDirectDrawDLL.own(LoadLibrarySystem32(L"ddraw.dll"));
-  if (!mDirectDrawDLL) {
-    ddraw.SetFailed(FeatureStatus::Unavailable,
-                    "DirectDraw not available on this computer",
-                    "FEATURE_FAILURE_DDRAW_LIB"_ns);
-    return;
-  }
-
-  sDirectDrawCreateExFn = (decltype(DirectDrawCreateEx)*)GetProcAddress(
-      mDirectDrawDLL, "DirectDrawCreateEx");
-  if (!sDirectDrawCreateExFn) {
-    ddraw.SetFailed(FeatureStatus::Unavailable,
-                    "DirectDraw not available on this computer",
-                    "FEATURE_FAILURE_DDRAW_LIB"_ns);
-    return;
-  }
-
-  HRESULT hr;
-  MOZ_SEH_TRY {
-    hr = sDirectDrawCreateExFn(nullptr, getter_AddRefs(mDirectDraw),
-                               IID_IDirectDraw7, nullptr);
-  }
-  MOZ_SEH_EXCEPT(EXCEPTION_EXECUTE_HANDLER) {
-    ddraw.SetFailed(FeatureStatus::Failed, "Failed to create DirectDraw",
-                    "FEATURE_FAILURE_DDRAW_LIB"_ns);
-    gfxCriticalNote << "DoesCreatingDirectDrawFailed";
-    return;
-  }
-  if (FAILED(hr)) {
-    ddraw.SetFailed(FeatureStatus::Failed, "Failed to create DirectDraw",
-                    "FEATURE_FAILURE_DDRAW_LIB"_ns);
-    gfxCriticalNote << "DoesCreatingDirectDrawFailed " << hexa(hr);
-    return;
-  }
-}
-
-IDirectDraw7* DeviceManagerDx::GetDirectDraw() { return mDirectDraw; }
-
 void DeviceManagerDx::GetCompositorDevices(
     RefPtr<ID3D11Device>* aOutDevice,
     RefPtr<layers::DeviceAttachmentsD3D11>* aOutAttachments) {
