@@ -8,7 +8,6 @@
 #define GFX_FONT_H
 
 #include <new>
-#include <utility>
 #include <functional>
 #include "PLDHashTable.h"
 #include "ThebesRLBoxTypes.h"
@@ -20,12 +19,12 @@
 #include "mozilla/FontPropertyTypes.h"
 #include "mozilla/HashTable.h"
 #include "mozilla/MemoryReporting.h"
-#include "mozilla/MruCache.h"
 #include "mozilla/Mutex.h"
 #include "mozilla/RefPtr.h"
 #include "mozilla/RWLock.h"
 #include "mozilla/TypedEnumBits.h"
 #include "mozilla/UniquePtr.h"
+#include "mozilla/gfx/FontPaletteCache.h"
 #include "mozilla/gfx/MatrixFwd.h"
 #include "mozilla/gfx/Point.h"
 #include "mozilla/gfx/2D.h"
@@ -50,6 +49,7 @@
 #include "gfxFontEntry.h"
 #include "gfxFontFeatures.h"
 
+class FontVisibilityProvider;
 class gfxContext;
 class gfxGraphiteShaper;
 class gfxHarfBuzzShaper;
@@ -97,9 +97,14 @@ struct gfxFontStyle {
   gfxFontStyle();
   gfxFontStyle(FontSlantStyle aStyle, FontWeight aWeight, FontStretch aStretch,
                gfxFloat aSize, const FontSizeAdjust& aSizeAdjust,
-               bool aSystemFont, bool aPrinterFont, bool aWeightSynthesis,
-               bool aStyleSynthesis, bool aSmallCapsSynthesis,
-               bool aPositionSynthesis, uint32_t aLanguageOverride);
+               bool aSystemFont, bool aPrinterFont,
+#ifdef XP_WIN
+               bool aAllowForceGDIClassic,
+#endif
+               bool aWeightSynthesis,
+               mozilla::StyleFontSynthesisStyle aStyleSynthesis,
+               bool aSmallCapsSynthesis, bool aPositionSynthesis,
+               mozilla::StyleFontLanguageOverride aLanguageOverride);
   // Features are composed of (1) features from style rules (2) features
   // from feature settings rules and (3) family-specific features.  (1) and
   // (3) are guaranteed to be mutually exclusive
@@ -145,11 +150,7 @@ struct gfxFontStyle {
   // but the font in use does not explicitly support this; the author can
   // use font-language-override to request the Serbian option in the font
   // in order to get correct glyph shapes.)
-  uint32_t languageOverride;
-
-  // The estimated background color behind the text. Enables a special
-  // rendering mode when NS_GET_A(.) > 0. Only used for text in the chrome.
-  nscolor fontSmoothingBackgroundColor;
+  mozilla::StyleFontLanguageOverride languageOverride;
 
   // The Font{Weight,Stretch,SlantStyle} fields are each a 16-bit type.
 
@@ -188,12 +189,16 @@ struct gfxFontStyle {
   // Say that this font is used for print or print preview.
   bool printerFont : 1;
 
+#ifdef XP_WIN
+  bool allowForceGDIClassic : 1;
+#endif
+
   // Used to imitate -webkit-font-smoothing: antialiased
   bool useGrayscaleAntialiasing : 1;
 
   // Whether synthetic styles are allowed (required, in the case of position)
   bool allowSyntheticWeight : 1;
-  bool allowSyntheticStyle : 1;
+  mozilla::StyleFontSynthesisStyle synthesisStyle : 2;
   bool allowSyntheticSmallCaps : 1;
   bool useSyntheticPosition : 1;
 
@@ -238,11 +243,14 @@ struct gfxFontStyle {
            (stretch == other.stretch) && (variantCaps == other.variantCaps) &&
            (variantSubSuper == other.variantSubSuper) &&
            (allowSyntheticWeight == other.allowSyntheticWeight) &&
-           (allowSyntheticStyle == other.allowSyntheticStyle) &&
+           (synthesisStyle == other.synthesisStyle) &&
            (allowSyntheticSmallCaps == other.allowSyntheticSmallCaps) &&
            (useSyntheticPosition == other.useSyntheticPosition) &&
            (systemFont == other.systemFont) &&
            (printerFont == other.printerFont) &&
+#ifdef XP_WIN
+           (allowForceGDIClassic == other.allowForceGDIClassic) &&
+#endif
            (useGrayscaleAntialiasing == other.useGrayscaleAntialiasing) &&
            (baselineOffset == other.baselineOffset) &&
            mozilla::NumbersAreBitwiseIdentical(sizeAdjust, other.sizeAdjust) &&
@@ -253,8 +261,7 @@ struct gfxFontStyle {
            (variationSettings == other.variationSettings) &&
            (languageOverride == other.languageOverride) &&
            mozilla::NumbersAreBitwiseIdentical(autoOpticalSize,
-                                               other.autoOpticalSize) &&
-           (fontSmoothingBackgroundColor == other.fontSmoothingBackgroundColor);
+                                               other.autoOpticalSize);
   }
 };
 
@@ -354,7 +361,7 @@ class gfxFontCache final
         mWordCacheExpirationTimer->InitWithNamedFuncCallback(
             WordCacheExpirationTimerCallback, this,
             SHAPED_WORD_TIMEOUT_SECONDS * 1000, nsITimer::TYPE_REPEATING_SLACK,
-            "gfxFontCache::WordCacheExpiration");
+            "gfxFontCache::WordCacheExpiration"_ns);
         mTimerRunning = true;
       }
     }
@@ -766,6 +773,8 @@ class gfxShapedText {
       FLAG_BREAK_TYPE_NONE = 0,
       FLAG_BREAK_TYPE_NORMAL = 1,
       FLAG_BREAK_TYPE_HYPHEN = 2,
+      // Allow break before this position if needed to avoid overflow:
+      FLAG_BREAK_TYPE_EMERGENCY_WRAP = 3,
 
       FLAG_CHAR_IS_SPACE = 0x10000000U,
 
@@ -890,7 +899,7 @@ class gfxShapedText {
     }
     // Returns FLAGS_CAN_BREAK_BEFORE if the setting changed, 0 otherwise
     uint32_t SetCanBreakBefore(uint8_t aCanBreakBefore) {
-      MOZ_ASSERT(aCanBreakBefore <= 2, "Bogus break-before value!");
+      MOZ_ASSERT(aCanBreakBefore <= 3, "Bogus break-flags value!");
       uint32_t breakMask = (uint32_t(aCanBreakBefore) << FLAGS_CAN_BREAK_SHIFT);
       uint32_t toggle = breakMask ^ (mValue & FLAGS_CAN_BREAK_BEFORE);
       mValue ^= toggle;
@@ -1041,8 +1050,8 @@ class gfxShapedText {
     return mDetailedGlyphs->Get(aCharIndex);
   }
 
-  void AdjustAdvancesForSyntheticBold(float aSynBoldOffset, uint32_t aOffset,
-                                      uint32_t aLength);
+  void ApplyTrackingToClusters(gfxFloat aTrackingAdjustment, uint32_t aOffset,
+                               uint32_t aLength);
 
   // Mark clusters in the CompressedGlyph records, starting at aOffset,
   // based on the Unicode properties of the text in aString.
@@ -1373,6 +1382,7 @@ class gfxShapedWord final : public gfxShapedText {
     memset(mCharGlyphsStorage, 0, aLength * sizeof(CompressedGlyph));
     uint8_t* text = reinterpret_cast<uint8_t*>(&mCharGlyphsStorage[aLength]);
     memcpy(text, aText, aLength * sizeof(uint8_t));
+    SetupClusterBoundaries(0, aText, aLength);
   }
 
   gfxShapedWord(const char16_t* aText, uint32_t aLength, Script aRunScript,
@@ -1528,9 +1538,7 @@ class gfxFont {
   // and therefore needs us to use a mask for text-shadow even when
   // we're not actually blurring.
   bool AlwaysNeedsMaskForShadow() const {
-    return mFontEntry->TryGetColorGlyphs() || mFontEntry->TryGetSVGData(this) ||
-           mFontEntry->HasFontTable(TRUETYPE_TAG('C', 'B', 'D', 'T')) ||
-           mFontEntry->HasFontTable(TRUETYPE_TAG('s', 'b', 'i', 'x'));
+    return mFontEntry->AlwaysNeedsMaskForShadow();
   }
 
   // whether a feature is supported by the font (limited to a small set
@@ -1642,13 +1650,11 @@ class gfxFont {
    * We let layout specify spacing on either side of any
    * character. We need to specify both before and after
    * spacing so that substring measurement can do the right things.
-   * These values are in appunits. They're always an integral number of
-   * appunits, but we specify them in floats in case very large spacing
-   * values are required.
+   * These values are in appunits.
    */
   struct Spacing {
-    gfxFloat mBefore;
-    gfxFloat mAfter;
+    nscoord mBefore;
+    nscoord mAfter;
   };
   /**
    * Metrics for a particular string
@@ -1808,7 +1814,7 @@ class gfxFont {
   }
 
   template <typename T>
-  bool InitFakeSmallCapsRun(nsPresContext* aPresContext,
+  bool InitFakeSmallCapsRun(FontVisibilityProvider* aFontVisibilityProvider,
                             DrawTarget* aDrawTarget, gfxTextRun* aTextRun,
                             const T* aText, uint32_t aOffset, uint32_t aLength,
                             FontMatchType aMatchType,
@@ -2217,6 +2223,10 @@ class gfxFont {
   gfxFontStyle mStyle;
   mutable gfxFloat mAdjustedSize;
 
+  // Tracking adjustment to be applied for CSS px size mCachedTrackingSize.
+  gfxFloat mTracking = 0.0;
+  gfxFloat mCachedTrackingSize = -1.0;
+
   // Conversion factor from font units to dev units; note that this may be
   // zero (in the degenerate case where mAdjustedSize has become zero).
   // This is OK because we only multiply by this factor, never divide.
@@ -2277,6 +2287,22 @@ class gfxFont {
                         const FontDrawParams& aFontParams,
                         const mozilla::gfx::Point& aPoint, uint32_t aGlyphId);
 
+  class ColorGlyphCache {
+   public:
+    ColorGlyphCache() = default;
+    ~ColorGlyphCache() = default;
+
+    void SetColors(mozilla::gfx::sRGBColor aCurrentColor,
+                   mozilla::gfx::FontPalette* aPalette);
+
+    mozilla::HashMap<uint32_t, RefPtr<mozilla::gfx::SourceSurface>> mCache;
+
+   private:
+    mozilla::gfx::sRGBColor mCurrentColor;
+    RefPtr<mozilla::gfx::FontPalette> mPalette;
+  };
+  mozilla::UniquePtr<ColorGlyphCache> mColorGlyphCache;
+
   // Subclasses can override to return true if the platform is able to render
   // COLR-font glyphs directly, instead of us painting the layers explicitly.
   // (Currently used only for COLR.v0 fonts on macOS.)
@@ -2300,6 +2326,10 @@ class gfxFont {
 // are dependent on the specific font, so they are set per GlyphRun.
 
 struct MOZ_STACK_CLASS TextRunDrawParams {
+  explicit TextRunDrawParams(mozilla::gfx::PaletteCache& aPaletteCache)
+      : paletteCache(aPaletteCache) {}
+
+  mozilla::gfx::PaletteCache& paletteCache;
   RefPtr<mozilla::gfx::DrawTarget> dt;
   gfxContext* context = nullptr;
   gfxFont::Spacing* spacing = nullptr;
@@ -2313,41 +2343,13 @@ struct MOZ_STACK_CLASS TextRunDrawParams {
   gfxPattern* textStrokePattern = nullptr;
   const mozilla::gfx::StrokeOptions* strokeOpts = nullptr;
   const mozilla::gfx::DrawOptions* drawOpts = nullptr;
-  const mozilla::gfx::FontPaletteValueSet* paletteValueSet = nullptr;
   nsAtom* fontPalette = nullptr;
   DrawMode drawMode = DrawMode::GLYPH_FILL;
   bool isVerticalRun = false;
   bool isRTL = false;
   bool paintSVGGlyphs = true;
   bool allowGDI = true;
-
-  // MRU cache of color-font palettes being used by fonts in the run. We cache
-  // these in the TextRunDrawParams so that we can avoid re-creating a new
-  // palette (which can be quite expensive) for each individual glyph run.
-  using CacheKey = const gfxFont*;
-
-  struct CacheData {
-    CacheKey mKey;
-    mozilla::UniquePtr<nsTArray<mozilla::gfx::sRGBColor>> mPalette;
-  };
-
-  class PaletteCache
-      : public mozilla::MruCache<CacheKey, CacheData, PaletteCache> {
-   public:
-    static mozilla::HashNumber Hash(const CacheKey& aKey) {
-      return mozilla::HashGeneric(aKey);
-    }
-    static bool Match(const CacheKey& aKey, const CacheData& aVal) {
-      return aVal.mKey == aKey;
-    }
-  };
-
-  PaletteCache mPaletteCache;
-
-  // Returns a pointer to a palette owned by the PaletteCache. This is only
-  // valid until the next call to GetPaletteFor (which might evict it) or
-  // until the TextRunDrawParams goes out of scope.
-  nsTArray<mozilla::gfx::sRGBColor>* GetPaletteFor(const gfxFont* aFont);
+  bool hasTextShadow = false;
 };
 
 struct MOZ_STACK_CLASS FontDrawParams {
@@ -2359,15 +2361,20 @@ struct MOZ_STACK_CLASS FontDrawParams {
   mozilla::gfx::DrawOptions drawOptions;
   gfxFloat advanceDirection;
   mozilla::gfx::sRGBColor currentColor;
-  nsTArray<mozilla::gfx::sRGBColor>* palette;  // owned by TextRunDrawParams
+  RefPtr<mozilla::gfx::FontPalette> palette;
   mozilla::gfx::Rect fontExtents;
   bool isVerticalFont;
   bool haveSVGGlyphs;
   bool haveColorGlyphs;
+  bool hasTextShadow;  // whether we're rendering with a text-shadow
 };
 
 struct MOZ_STACK_CLASS EmphasisMarkDrawParams {
+  EmphasisMarkDrawParams(gfxContext* aContext,
+                         mozilla::gfx::PaletteCache& aPaletteCache)
+      : context(aContext), paletteCache(aPaletteCache) {}
   gfxContext* context;
+  mozilla::gfx::PaletteCache& paletteCache;
   gfxFont::Spacing* spacing;
   gfxTextRun* mark;
   gfxFloat advance;
