@@ -9,17 +9,14 @@
 #include <string.h>
 #include <algorithm>
 #include <new>
-#include <utility>
 #include "gfxPlatform.h"
-#include "ipc/IPCMessageUtils.h"
-#include "ipc/IPCMessageUtilsSpecializations.h"
+#include "harfbuzz/hb.h"
 #include "mozilla/Assertions.h"
 #include "mozilla/Attributes.h"
 #include "mozilla/Casting.h"
 #include "mozilla/EndianUtils.h"
-#include "mozilla/FontPropertyTypes.h"
+#include "mozilla/ServoStyleConstsInlines.h"
 #include "mozilla/MemoryReporting.h"
-#include "mozilla/UniquePtr.h"
 #include "nsStringFwd.h"
 #include "nsTArray.h"
 #include "nscore.h"
@@ -32,9 +29,7 @@ struct gfxFontVariationInstance;
 
 namespace mozilla {
 class Encoding;
-namespace gfx {
-struct DeviceColor;
-}
+class ServoStyleSet;
 }  // namespace mozilla
 
 /* Bug 341128 - w32api defines min/max which causes problems with <bitset> */
@@ -51,6 +46,11 @@ typedef struct hb_blob_t hb_blob_t;
 
 class SharedBitSet;
 
+namespace IPC {
+template <typename T>
+struct ParamTraits;
+}
+
 class gfxSparseBitSet {
  private:
   friend class SharedBitSet;
@@ -59,15 +59,34 @@ class gfxSparseBitSet {
   enum { BLOCK_SIZE_BITS = BLOCK_SIZE * 8 };
   enum { NO_BLOCK = 0xffff };  // index value indicating missing (empty) block
 
+  // The BlockIndex type is just a uint16_t, except it will default-construct
+  // with the value NO_BLOCK so that we can do AppendElements(size_t) to grow
+  // the index.
+  struct BlockIndex {
+    BlockIndex() : mIndex(NO_BLOCK) {}
+    explicit BlockIndex(uint16_t aIndex) : mIndex(aIndex) {}
+
+    operator uint16_t() const { return mIndex; }
+
+    uint16_t mIndex;
+  };
+
   struct Block {
-    explicit Block(unsigned char memsetValue = 0) {
+    Block() { memset(mBits, 0, BLOCK_SIZE); }
+    explicit Block(unsigned char memsetValue) {
       memset(mBits, memsetValue, BLOCK_SIZE);
     }
     uint8_t mBits[BLOCK_SIZE];
   };
 
+  friend struct IPC::ParamTraits<gfxSparseBitSet>;
+  friend struct IPC::ParamTraits<BlockIndex>;
+  friend struct IPC::ParamTraits<Block>;
+
  public:
   gfxSparseBitSet() = default;
+  explicit gfxSparseBitSet(uint32_t aReserveCapacity)
+      : mBlockIndex(aReserveCapacity), mBlocks(aReserveCapacity) {}
 
   bool Equals(const gfxSparseBitSet* aOther) const {
     if (mBlockIndex.Length() != aOther->mBlockIndex.Length()) {
@@ -75,8 +94,8 @@ class gfxSparseBitSet {
     }
     size_t n = mBlockIndex.Length();
     for (size_t i = 0; i < n; ++i) {
-      uint32_t b1 = mBlockIndex[i];
-      uint32_t b2 = aOther->mBlockIndex[i];
+      uint16_t b1 = mBlockIndex[i];
+      uint16_t b2 = aOther->mBlockIndex[i];
       if ((b1 == NO_BLOCK) != (b2 == NO_BLOCK)) {
         return false;
       }
@@ -171,13 +190,13 @@ class gfxSparseBitSet {
 
   void set(uint32_t aIndex) {
     uint32_t i = aIndex / BLOCK_SIZE_BITS;
-    while (i >= mBlockIndex.Length()) {
-      mBlockIndex.AppendElement(NO_BLOCK);
+    if (i >= mBlockIndex.Length()) {
+      mBlockIndex.AppendElements(i - mBlockIndex.Length() + 1);
     }
     if (mBlockIndex[i] == NO_BLOCK) {
       mBlocks.AppendElement();
       MOZ_ASSERT(mBlocks.Length() < 0xffff, "block index overflow!");
-      mBlockIndex[i] = static_cast<uint16_t>(mBlocks.Length() - 1);
+      mBlockIndex[i].mIndex = static_cast<uint16_t>(mBlocks.Length() - 1);
     }
     Block& block = mBlocks[mBlockIndex[i]];
     block.mBits[(aIndex >> 3) & (BLOCK_SIZE - 1)] |= 1 << (aIndex & 0x7);
@@ -195,8 +214,8 @@ class gfxSparseBitSet {
     const uint32_t startIndex = aStart / BLOCK_SIZE_BITS;
     const uint32_t endIndex = aEnd / BLOCK_SIZE_BITS;
 
-    while (endIndex >= mBlockIndex.Length()) {
-      mBlockIndex.AppendElement(NO_BLOCK);
+    if (endIndex >= mBlockIndex.Length()) {
+      mBlockIndex.AppendElements(endIndex - mBlockIndex.Length() + 1);
     }
 
     for (uint32_t i = startIndex; i <= endIndex; ++i) {
@@ -205,9 +224,9 @@ class gfxSparseBitSet {
 
       if (mBlockIndex[i] == NO_BLOCK) {
         bool fullBlock = (aStart <= blockFirstBit && aEnd >= blockLastBit);
-        mBlocks.AppendElement(Block(fullBlock ? 0xFF : 0));
+        mBlocks.AppendElement(fullBlock ? Block(0xFF) : Block());
         MOZ_ASSERT(mBlocks.Length() < 0xffff, "block index overflow!");
-        mBlockIndex[i] = static_cast<uint16_t>(mBlocks.Length() - 1);
+        mBlockIndex[i].mIndex = static_cast<uint16_t>(mBlocks.Length() - 1);
         if (fullBlock) {
           continue;
         }
@@ -231,9 +250,7 @@ class gfxSparseBitSet {
       return;
     }
     if (mBlockIndex[i] == NO_BLOCK) {
-      mBlocks.AppendElement();
-      MOZ_ASSERT(mBlocks.Length() < 0xffff, "block index overflow!");
-      mBlockIndex[i] = static_cast<uint16_t>(mBlocks.Length() - 1);
+      return;
     }
     Block& block = mBlocks[mBlockIndex[i]];
     block.mBits[(aIndex >> 3) & (BLOCK_SIZE - 1)] &= ~(1 << (aIndex & 0x7));
@@ -282,10 +299,10 @@ class gfxSparseBitSet {
 
   // set this bitset to the union of its current contents and another
   void Union(const gfxSparseBitSet& aBitset) {
-    // ensure mBlocks is large enough
+    // ensure mBlockIndex is large enough
     uint32_t blockCount = aBitset.mBlockIndex.Length();
-    while (blockCount > mBlockIndex.Length()) {
-      mBlockIndex.AppendElement(NO_BLOCK);
+    if (blockCount > mBlockIndex.Length()) {
+      mBlockIndex.AppendElements(blockCount - mBlockIndex.Length());
     }
     // for each block that may be present in aBitset...
     for (uint32_t i = 0; i < blockCount; ++i) {
@@ -297,7 +314,7 @@ class gfxSparseBitSet {
       if (mBlockIndex[i] == NO_BLOCK) {
         mBlocks.AppendElement(aBitset.mBlocks[aBitset.mBlockIndex[i]]);
         MOZ_ASSERT(mBlocks.Length() < 0xffff, "block index overflow!");
-        mBlockIndex[i] = static_cast<uint16_t>(mBlocks.Length() - 1);
+        mBlockIndex[i].mIndex = static_cast<uint16_t>(mBlocks.Length() - 1);
         continue;
       }
       // else set existing block to the union of both
@@ -330,41 +347,11 @@ class gfxSparseBitSet {
     return check;
   }
 
- private:
-  friend struct IPC::ParamTraits<gfxSparseBitSet>;
-  friend struct IPC::ParamTraits<gfxSparseBitSet::Block>;
-  CopyableTArray<uint16_t> mBlockIndex;
+ protected:
+  CopyableTArray<BlockIndex> mBlockIndex;
   CopyableTArray<Block> mBlocks;
 };
 
-namespace IPC {
-template <>
-struct ParamTraits<gfxSparseBitSet> {
-  typedef gfxSparseBitSet paramType;
-  static void Write(Message* aMsg, const paramType& aParam) {
-    WriteParam(aMsg, aParam.mBlockIndex);
-    WriteParam(aMsg, aParam.mBlocks);
-  }
-  static bool Read(const Message* aMsg, PickleIterator* aIter,
-                   paramType* aResult) {
-    return ReadParam(aMsg, aIter, &aResult->mBlockIndex) &&
-           ReadParam(aMsg, aIter, &aResult->mBlocks);
-  }
-};
-
-template <>
-struct ParamTraits<gfxSparseBitSet::Block> {
-  typedef gfxSparseBitSet::Block paramType;
-  static void Write(Message* aMsg, const paramType& aParam) {
-    aMsg->WriteBytes(&aParam, sizeof(aParam));
-  }
-  static bool Read(const Message* aMsg, PickleIterator* aIter,
-                   paramType* aResult) {
-    return aMsg->ReadBytesInto(aIter, aResult, sizeof(*aResult));
-  }
-};
-}  // namespace IPC
-
 /**
  * SharedBitSet is a version of gfxSparseBitSet that is intended to be used
  * in a shared-memory block, and can be used regardless of the address at which
@@ -509,14 +496,17 @@ inline void gfxSparseBitSet::Union(const SharedBitSet& aBitset) {
     if (mBlockIndex[i] == NO_BLOCK) {
       mBlocks.AppendElement(blocks[blockIndex[i]]);
       MOZ_ASSERT(mBlocks.Length() < 0xffff, "block index overflow");
-      mBlockIndex[i] = uint16_t(mBlocks.Length() - 1);
+      mBlockIndex[i].mIndex = uint16_t(mBlocks.Length() - 1);
       continue;
     }
-    // else set existing target block to the union of both
-    uint32_t* dst = reinterpret_cast<uint32_t*>(&mBlocks[mBlockIndex[i]].mBits);
-    const uint32_t* src =
-        reinterpret_cast<const uint32_t*>(&blocks[blockIndex[i]].mBits);
-    for (uint32_t j = 0; j < BLOCK_SIZE / 4; ++j) {
+    // Else set existing target block to the union of both.
+    // Note that blocks in SharedBitSet may not be 4-byte aligned, so we don't
+    // try to optimize by casting to uint32_t* here and processing 4 bytes at
+    // once, as this could result in misaligned access.
+    uint8_t* dst = reinterpret_cast<uint8_t*>(&mBlocks[mBlockIndex[i]].mBits);
+    const uint8_t* src =
+        reinterpret_cast<const uint8_t*>(&blocks[blockIndex[i]].mBits);
+    for (uint32_t j = 0; j < BLOCK_SIZE; ++j) {
       dst[j] |= src[j];
     }
   }
@@ -810,22 +800,6 @@ struct KernTableSubtableHeaderVersion1 {
   AutoSwap_PRUint16 tupleIndex;
 };
 
-struct COLRHeader {
-  AutoSwap_PRUint16 version;
-  AutoSwap_PRUint16 numBaseGlyphRecord;
-  AutoSwap_PRUint32 offsetBaseGlyphRecord;
-  AutoSwap_PRUint32 offsetLayerRecord;
-  AutoSwap_PRUint16 numLayerRecords;
-};
-
-struct CPALHeaderVersion0 {
-  AutoSwap_PRUint16 version;
-  AutoSwap_PRUint16 numPaletteEntries;
-  AutoSwap_PRUint16 numPalettes;
-  AutoSwap_PRUint16 numColorRecords;
-  AutoSwap_PRUint32 offsetFirstColorRecord;
-};
-
 #pragma pack()
 
 // Return just the highest bit of the given value, i.e., the highest
@@ -947,6 +921,19 @@ class gfxFontUtils {
                                         // (in bytes).
   };
 
+  // Helper to ensure we free a font table when we return.
+  class AutoHBBlob {
+   public:
+    explicit AutoHBBlob(hb_blob_t* aBlob) : mBlob(aBlob) {}
+
+    ~AutoHBBlob() { hb_blob_destroy(mBlob); }
+
+    operator hb_blob_t*() { return mBlob; }
+
+   private:
+    hb_blob_t* const mBlob;
+  };
+
   // for reading big-endian font data on either big or little-endian platforms
 
   static inline uint16_t ReadShortAt(const uint8_t* aBuf, uint32_t aIndex) {
@@ -977,15 +964,17 @@ class gfxFontUtils {
                                             gfxSparseBitSet& aCharacterMap);
 
   static nsresult ReadCMAPTableFormat4(const uint8_t* aBuf, uint32_t aLength,
-                                       gfxSparseBitSet& aCharacterMap);
+                                       gfxSparseBitSet& aCharacterMap,
+                                       bool aIsSymbolFont);
 
   static nsresult ReadCMAPTableFormat14(const uint8_t* aBuf, uint32_t aLength,
-                                        mozilla::UniquePtr<uint8_t[]>& aTable);
+                                        const uint8_t*& aTable);
 
   static uint32_t FindPreferredSubtable(const uint8_t* aBuf,
                                         uint32_t aBufLength,
                                         uint32_t* aTableOffset,
-                                        uint32_t* aUVSTableOffset);
+                                        uint32_t* aUVSTableOffset,
+                                        bool* aIsSymbolFont);
 
   static nsresult ReadCMAP(const uint8_t* aBuf, uint32_t aBufLength,
                            gfxSparseBitSet& aCharacterMap,
@@ -1001,6 +990,10 @@ class gfxFontUtils {
   static uint16_t MapUVSToGlyphFormat14(const uint8_t* aBuf, uint32_t aCh,
                                         uint32_t aVS);
 
+  // Return whether <aCh, aVS> is supported as the default variation for aCh.
+  static bool IsDefaultUVSSequence(const uint8_t* aBuf, uint32_t aCh,
+                                   uint32_t aVS);
+
   // sCJKCompatSVSTable is a 'cmap' format 14 subtable that maps
   // <char + var-selector> pairs to the corresponding Unicode
   // compatibility ideograph codepoints.
@@ -1012,6 +1005,12 @@ class gfxFontUtils {
   static uint32_t MapCharToGlyph(const uint8_t* aCmapBuf, uint32_t aBufLength,
                                  uint32_t aUnicode, uint32_t aVarSelector = 0);
 
+  // For legacy MS Symbol fonts, we try mapping 8-bit character codes to the
+  // Private Use range at U+F0xx used by the cmaps in these fonts.
+  static MOZ_ALWAYS_INLINE uint32_t MapLegacySymbolFontCharToPUA(uint32_t aCh) {
+    return aCh >= 0x20 && aCh <= 0xff ? 0xf000 + aCh : 0;
+  }
+
 #ifdef XP_WIN
   // determine whether a font (which has already been sanitized, so is known
   // to be a valid sfnt) is CFF format rather than TrueType
@@ -1110,44 +1109,12 @@ class gfxFontUtils {
            aCh <= kUnicodeRegionalIndicatorZ;
   }
 
-  static inline bool IsInvalid(uint32_t ch) { return (ch == 0xFFFD); }
+  static inline bool IsEmojiFlagAndTag(uint32_t aCh, uint32_t aNext) {
+    constexpr uint32_t kBlackFlag = 0x1F3F4;
+    constexpr uint32_t kTagLetterA = 0xE0061;
+    constexpr uint32_t kTagLetterZ = 0xE007A;
 
-  // Font code may want to know if there is the potential for bidi behavior
-  // to be triggered by any of the characters in a text run; this can be
-  // used to test that possibility.
-  enum {
-    kUnicodeBidiScriptsStart = 0x0590,
-    kUnicodeBidiScriptsEnd = 0x08FF,
-    kUnicodeBidiPresentationStart = 0xFB1D,
-    kUnicodeBidiPresentationEnd = 0xFEFC,
-    kUnicodeFirstHighSurrogateBlock = 0xD800,
-    kUnicodeRLM = 0x200F,
-    kUnicodeRLE = 0x202B,
-    kUnicodeRLO = 0x202E
-  };
-
-  static inline bool PotentialRTLChar(char16_t aCh) {
-    if (aCh >= kUnicodeBidiScriptsStart && aCh <= kUnicodeBidiScriptsEnd)
-      // bidi scripts Hebrew, Arabic, Syriac, Thaana, N'Ko are all encoded
-      // together
-      return true;
-
-    if (aCh == kUnicodeRLM || aCh == kUnicodeRLE || aCh == kUnicodeRLO)
-      // directional controls that trigger bidi layout
-      return true;
-
-    if (aCh >= kUnicodeBidiPresentationStart &&
-        aCh <= kUnicodeBidiPresentationEnd)
-      // presentation forms of Arabic and Hebrew letters
-      return true;
-
-    if ((aCh & 0xFF00) == kUnicodeFirstHighSurrogateBlock)
-      // surrogate that could be part of a bidi supplementary char
-      // (Cypriot, Aramaic, Phoenecian, etc)
-      return true;
-
-    // otherwise we know this char cannot trigger bidi reordering
-    return false;
+    return aCh == kBlackFlag && aNext >= kTagLetterA && aNext <= kTagLetterZ;
   }
 
   // parse a simple list of font family names into
@@ -1155,28 +1122,13 @@ class gfxFontUtils {
   static void ParseFontList(const nsACString& aFamilyList,
                             nsTArray<nsCString>& aFontList);
 
-  // for a given font list pref name, append list of font names
-  static void AppendPrefsFontList(const char* aPrefName,
-                                  nsTArray<nsCString>& aFontList,
-                                  bool aLocalized = false);
-
-  // for a given font list pref name, initialize a list of font names
+  // for a given pref name, initialize a list of font names
   static void GetPrefsFontList(const char* aPrefName,
-                               nsTArray<nsCString>& aFontList,
-                               bool aLocalized = false);
+                               nsTArray<nsCString>& aFontList);
 
   // generate a unique font name
   static nsresult MakeUniqueUserFontName(nsAString& aName);
 
-  // for color layer from glyph using COLR and CPAL tables
-  static bool ValidateColorGlyphs(hb_blob_t* aCOLR, hb_blob_t* aCPAL);
-  static bool GetColorGlyphLayers(
-      hb_blob_t* aCOLR, hb_blob_t* aCPAL, uint32_t aGlyphId,
-      const mozilla::gfx::DeviceColor& aDefaultColor,
-      nsTArray<uint16_t>& aGlyphs,
-      nsTArray<mozilla::gfx::DeviceColor>& aColors);
-  static bool HasColorLayersForGlyph(hb_blob_t* aCOLR, uint32_t aGlyphId);
-
   // Helper used to implement gfxFontEntry::GetVariation{Axes,Instances} for
   // platforms where the native font APIs don't provide the info we want
   // in a convenient form, or when native APIs are too expensive.
@@ -1193,6 +1145,22 @@ class gfxFontUtils {
       uint32_t aDataLength, nsTArray<nsCString>& aOtherFamilyNames,
       bool useFullName);
 
+  // Main, DOM worker or servo thread safe method to check if we are performing
+  // Servo traversal.
+  static bool IsInServoTraversal();
+
+  // Main, DOM worker or servo thread safe method to get the current
+  // ServoTypeSet. Always returns nullptr for DOM worker threads.
+  static mozilla::ServoStyleSet* CurrentServoStyleSet();
+
+  static void AssertSafeThreadOrServoFontMetricsLocked()
+#ifdef DEBUG
+      ;
+#else
+  {
+  }
+#endif
+
  protected:
   friend struct MacCharsetMappingComparator;
 
@@ -1231,11 +1199,15 @@ constexpr double kStretchFactor = 1.0e8;
 constexpr double kStyleFactor = 1.0e4;
 constexpr double kWeightFactor = 1.0e0;
 
-// style distance ==> [0,500]
+// If the output range of this function is extended, check assertions &
+// usage at the callsites!
+// style distance ==> [0,900]
 static inline double StyleDistance(const mozilla::SlantStyleRange& aRange,
-                                   mozilla::FontSlantStyle aTargetStyle) {
+                                   mozilla::FontSlantStyle aTargetStyle,
+                                   bool aItalicToObliqueFallback) {
   const mozilla::FontSlantStyle minStyle = aRange.Min();
-  if (aTargetStyle == minStyle) {
+  const mozilla::FontSlantStyle maxStyle = aRange.Max();
+  if (aTargetStyle == minStyle || aTargetStyle == maxStyle) {
     return 0.0;  // styles match exactly ==> 0
   }
 
@@ -1247,146 +1219,130 @@ static inline double StyleDistance(const mozilla::SlantStyleRange& aRange,
   // vice versa
   const double kNegate = 200.0;
 
+  // bias added for oblique faces when italic is requested, and only-oblique
+  // font-synthesis-style is in effect
+  const double kBadFallback = 400.0;
+
   if (aTargetStyle.IsNormal()) {
-    if (minStyle.IsOblique()) {
-      // to distinguish oblique 0deg from normal, we add 1.0 to the angle
-      const double minAngle = minStyle.ObliqueAngle();
-      if (minAngle >= 0.0) {
-        return 1.0 + minAngle;
-      }
-      const mozilla::FontSlantStyle maxStyle = aRange.Max();
-      const double maxAngle = maxStyle.ObliqueAngle();
-      if (maxAngle >= 0.0) {
-        // [min,max] range includes 0.0, so just return our minimum
-        return 1.0;
-      }
-      // negative oblique is even worse than italic
-      return kNegate - maxAngle;
+    if (minStyle.IsItalic() || maxStyle.IsItalic()) {
+      // italic is the worst match (prefer any oblique; 0deg was requested)
+      return kBadFallback + kNegate + kReverse;
     }
-    // must be italic, which is worse than any non-negative oblique;
-    // treat as a match in the wrong search direction
-    MOZ_ASSERT(minStyle.IsItalic());
-    return kReverse;
+    const double minAngle = minStyle.ObliqueAngle();
+    if (minAngle >= 0.0) {
+      return minAngle;
+    }
+    const double maxAngle = maxStyle.ObliqueAngle();
+    if (maxAngle >= 0.0) {
+      // [min,max] range includes 0.0, so it's a perfect match
+      return 0.0;
+    }
+    return kNegate - maxAngle;
   }
 
-  const double kDefaultAngle =
-      mozilla::FontSlantStyle::Oblique().ObliqueAngle();
+  const double kDefaultAngle = mozilla::FontSlantStyle::DEFAULT_OBLIQUE_DEGREES;
 
   if (aTargetStyle.IsItalic()) {
-    if (minStyle.IsOblique()) {
-      const double minAngle = minStyle.ObliqueAngle();
-      if (minAngle >= kDefaultAngle) {
-        return 1.0 + (minAngle - kDefaultAngle);
-      }
-      const mozilla::FontSlantStyle maxStyle = aRange.Max();
-      const double maxAngle = maxStyle.ObliqueAngle();
-      if (maxAngle >= kDefaultAngle) {
-        return 1.0;
-      }
-      if (maxAngle > 0.0) {
-        // wrong direction but still > 0, add bias of 100
-        return kReverse + (kDefaultAngle - maxAngle);
-      }
-      // negative oblique angle, add bias of 300
-      return kReverse + kNegate + (kDefaultAngle - maxAngle);
+    MOZ_ASSERT(!minStyle.IsItalic());  // we checked for equality above
+    double targetAngle = kDefaultAngle;
+    double fallbackBias = 0.0;
+    if (!aItalicToObliqueFallback) {
+      // If 'font-style-synthesis: oblique-only' is applied, we should not use
+      // oblique as a fallback for italic, so we add a large "fallback bias" to
+      // all results here, and prefer an angle as close to zero as possible.
+      targetAngle = 0.0;
+      fallbackBias = kBadFallback;
     }
-    // normal is worse than oblique > 0, but better than oblique <= 0
-    MOZ_ASSERT(minStyle.IsNormal());
-    return kNegate;
+    const double minAngle = minStyle.ObliqueAngle();
+    if (minAngle >= targetAngle) {
+      // Add 1.0 to ensure italic vs non-italic never returns 0.0, even if the
+      // angle matches.
+      return fallbackBias + minAngle - targetAngle + 1.0;
+    }
+    const double maxAngle = maxStyle.ObliqueAngle();
+    if (maxAngle >= targetAngle) {
+      return fallbackBias + 1.0;
+    }
+    if (maxAngle > 0.0) {
+      // wrong direction but still > 0, add bias of 100
+      return fallbackBias + kReverse + (targetAngle - maxAngle);
+    }
+    // negative oblique angle, add bias of 300
+    return fallbackBias + kReverse + kNegate + (targetAngle - maxAngle);
   }
 
   // target is oblique <angle>: four different cases depending on
   // the value of the <angle>, which determines the preferred direction
   // of search
   const double targetAngle = aTargetStyle.ObliqueAngle();
+
+  // italic is a bad fallback if it was not requested
+  if (minStyle.IsItalic() || maxStyle.IsItalic()) {
+    return kBadFallback + kNegate + kReverse;
+  }
+
   if (targetAngle >= kDefaultAngle) {
-    if (minStyle.IsOblique()) {
-      const double minAngle = minStyle.ObliqueAngle();
-      if (minAngle >= targetAngle) {
-        return minAngle - targetAngle;
-      }
-      const mozilla::FontSlantStyle maxStyle = aRange.Max();
-      const double maxAngle = maxStyle.ObliqueAngle();
-      if (maxAngle >= targetAngle) {
-        return 0.0;
-      }
-      if (maxAngle > 0.0) {
-        return kReverse + (targetAngle - maxAngle);
-      }
-      return kReverse + kNegate + (targetAngle - maxAngle);
+    const double minAngle = minStyle.ObliqueAngle();
+    if (minAngle >= targetAngle) {
+      return minAngle - targetAngle;
     }
-    if (minStyle.IsItalic()) {
-      return kReverse + kNegate;
+    const double maxAngle = maxStyle.ObliqueAngle();
+    if (maxAngle >= targetAngle) {
+      return 0.0;
     }
-    return kReverse + kNegate + 1.0;
+    if (maxAngle > 0.0) {
+      return kReverse + (targetAngle - maxAngle);
+    }
+    return kReverse + kNegate + (targetAngle - maxAngle);
   }
 
   if (targetAngle <= -kDefaultAngle) {
-    if (minStyle.IsOblique()) {
-      const mozilla::FontSlantStyle maxStyle = aRange.Max();
-      const double maxAngle = maxStyle.ObliqueAngle();
-      if (maxAngle <= targetAngle) {
-        return targetAngle - maxAngle;
-      }
-      const double minAngle = minStyle.ObliqueAngle();
-      if (minAngle <= targetAngle) {
-        return 0.0;
-      }
-      if (minAngle < 0.0) {
-        return kReverse + (minAngle - targetAngle);
-      }
-      return kReverse + kNegate + (minAngle - targetAngle);
-    }
-    if (minStyle.IsItalic()) {
-      return kReverse + kNegate;
-    }
-    return kReverse + kNegate + 1.0;
-  }
-
-  if (targetAngle >= 0.0) {
-    if (minStyle.IsOblique()) {
-      const double minAngle = minStyle.ObliqueAngle();
-      if (minAngle > targetAngle) {
-        return kReverse + (minAngle - targetAngle);
-      }
-      const mozilla::FontSlantStyle maxStyle = aRange.Max();
-      const double maxAngle = maxStyle.ObliqueAngle();
-      if (maxAngle >= targetAngle) {
-        return 0.0;
-      }
-      if (maxAngle > 0.0) {
-        return targetAngle - maxAngle;
-      }
-      return kReverse + kNegate + (targetAngle - maxAngle);
-    }
-    if (minStyle.IsItalic()) {
-      return kReverse + kNegate - 2.0;
-    }
-    return kReverse + kNegate - 1.0;
-  }
-
-  // last case: (targetAngle < 0.0 && targetAngle > kDefaultAngle)
-  if (minStyle.IsOblique()) {
-    const mozilla::FontSlantStyle maxStyle = aRange.Max();
     const double maxAngle = maxStyle.ObliqueAngle();
-    if (maxAngle < targetAngle) {
-      return kReverse + (targetAngle - maxAngle);
+    if (maxAngle <= targetAngle) {
+      return targetAngle - maxAngle;
     }
     const double minAngle = minStyle.ObliqueAngle();
     if (minAngle <= targetAngle) {
       return 0.0;
     }
     if (minAngle < 0.0) {
-      return minAngle - targetAngle;
+      return kReverse + (minAngle - targetAngle);
     }
     return kReverse + kNegate + (minAngle - targetAngle);
   }
-  if (minStyle.IsItalic()) {
-    return kReverse + kNegate - 2.0;
+
+  if (targetAngle >= 0.0) {
+    const double minAngle = minStyle.ObliqueAngle();
+    if (minAngle > targetAngle) {
+      return kReverse + (minAngle - targetAngle);
+    }
+    const double maxAngle = maxStyle.ObliqueAngle();
+    if (maxAngle >= targetAngle) {
+      return 0.0;
+    }
+    if (maxAngle > 0.0) {
+      return targetAngle - maxAngle;
+    }
+    return kReverse + kNegate + (targetAngle - maxAngle);
   }
-  return kReverse + kNegate - 1.0;
+
+  // last case: (targetAngle < 0.0 && targetAngle > kDefaultAngle)
+  const double maxAngle = maxStyle.ObliqueAngle();
+  if (maxAngle < targetAngle) {
+    return kReverse + (targetAngle - maxAngle);
+  }
+  const double minAngle = minStyle.ObliqueAngle();
+  if (minAngle <= targetAngle) {
+    return 0.0;
+  }
+  if (minAngle < 0.0) {
+    return minAngle - targetAngle;
+  }
+  return kReverse + kNegate + (minAngle - targetAngle);
 }
 
+// If the output range of this function is extended, check assertions &
+// usage at the callsites!
 // stretch distance ==> [0,2000]
 static inline double StretchDistance(const mozilla::StretchRange& aRange,
                                      mozilla::FontStretch aTargetStretch) {
@@ -1401,16 +1357,16 @@ static inline double StretchDistance(const mozilla::StretchRange& aRange,
   // If aTargetStretch is >100, we prefer larger values if available;
   // if <=100, we prefer smaller values if available.
   if (aTargetStretch < minStretch) {
-    if (aTargetStretch > mozilla::FontStretch::Normal()) {
-      return minStretch - aTargetStretch;
+    if (aTargetStretch > mozilla::FontStretch::NORMAL) {
+      return minStretch.ToFloat() - aTargetStretch.ToFloat();
     }
-    return (minStretch - aTargetStretch) + kReverseDistance;
+    return (minStretch.ToFloat() - aTargetStretch.ToFloat()) + kReverseDistance;
   }
   if (aTargetStretch > maxStretch) {
-    if (aTargetStretch <= mozilla::FontStretch::Normal()) {
-      return aTargetStretch - maxStretch;
+    if (aTargetStretch <= mozilla::FontStretch::NORMAL) {
+      return aTargetStretch.ToFloat() - maxStretch.ToFloat();
     }
-    return (aTargetStretch - maxStretch) + kReverseDistance;
+    return (aTargetStretch.ToFloat() - maxStretch.ToFloat()) + kReverseDistance;
   }
   return 0.0;
 }
@@ -1426,6 +1382,8 @@ static inline double StretchDistance(const mozilla::StretchRange& aRange,
 // weights are farther away than lighter weights. If the target is 5 and the
 // font weight 995, the distance would be 1590 for the same reason.
 
+// If the output range of this function is extended, check assertions &
+// usage at the callsites!
 // weight distance ==> [0,1600]
 static inline double WeightDistance(const mozilla::WeightRange& aRange,
                                     mozilla::FontWeight aTargetWeight) {
@@ -1440,35 +1398,36 @@ static inline double WeightDistance(const mozilla::WeightRange& aRange,
     return 0.0;
   }
 
-  if (aTargetWeight < mozilla::FontWeight(400)) {
+  if (aTargetWeight < mozilla::FontWeight::NORMAL) {
     // Requested a lighter-than-400 weight
     if (maxWeight < aTargetWeight) {
-      return aTargetWeight - maxWeight;
+      return aTargetWeight.ToFloat() - maxWeight.ToFloat();
     }
     // Add reverse-search penalty for bolder faces
-    return (minWeight - aTargetWeight) + kReverseDistance;
+    return (minWeight.ToFloat() - aTargetWeight.ToFloat()) + kReverseDistance;
   }
 
-  if (aTargetWeight > mozilla::FontWeight(500)) {
+  if (aTargetWeight > mozilla::FontWeight::FromInt(500)) {
     // Requested a bolder-than-500 weight
     if (minWeight > aTargetWeight) {
-      return minWeight - aTargetWeight;
+      return minWeight.ToFloat() - aTargetWeight.ToFloat();
     }
     // Add reverse-search penalty for lighter faces
-    return (aTargetWeight - maxWeight) + kReverseDistance;
+    return (aTargetWeight.ToFloat() - maxWeight.ToFloat()) + kReverseDistance;
   }
 
   // Special case for requested weight in the [400..500] range
   if (minWeight > aTargetWeight) {
-    if (minWeight <= mozilla::FontWeight(500)) {
+    if (minWeight <= mozilla::FontWeight::FromInt(500)) {
       // Bolder weight up to 500 is first choice
-      return minWeight - aTargetWeight;
+      return minWeight.ToFloat() - aTargetWeight.ToFloat();
     }
     // Other bolder weights get a reverse-search penalty
-    return (minWeight - aTargetWeight) + kReverseDistance;
+    return (minWeight.ToFloat() - aTargetWeight.ToFloat()) + kReverseDistance;
   }
   // Lighter weights are not as good as bolder ones within [400..500]
-  return (aTargetWeight - maxWeight) + kNotWithinCentralRange;
+  return (aTargetWeight.ToFloat() - maxWeight.ToFloat()) +
+         kNotWithinCentralRange;
 }
 
 #endif /* GFX_FONT_UTILS_H */
