@@ -19,7 +19,6 @@
 #include "nsTArray.h"
 
 #include "mozilla/Mutex.h"
-#include "mozilla/RefPtr.h"
 
 #include <windows.h>
 #include <objbase.h>
@@ -103,51 +102,20 @@ class gfxWindowsPlatform final : public gfxPlatform {
   virtual already_AddRefed<gfxASurface> CreateOffscreenSurface(
       const IntSize& aSize, gfxImageFormat aFormat) override;
 
-  enum RenderMode {
-    /* Use GDI and windows surfaces */
-    RENDER_GDI = 0,
-
-    /* Use 32bpp image surfaces and call StretchDIBits */
-    RENDER_IMAGE_STRETCH32,
-
-    /* Use 32bpp image surfaces, and do 32->24 conversion before calling
-       StretchDIBits */
-    RENDER_IMAGE_STRETCH24,
-
-    /* Use Direct2D rendering */
-    RENDER_DIRECT2D,
-
-    /* max */
-    RENDER_MODE_MAX
-  };
-
-  bool IsDirect2DBackend();
-
   /**
    * Updates render mode with relation to the current preferences and
    * available devices.
    */
   void UpdateRenderMode();
 
-  /**
-   * Verifies a D2D device is present and working, will attempt to create one
-   * it is non-functional or non-existant.
-   *
-   * \param aAttemptForce Attempt to force D2D cairo device creation by using
-   * cairo device creation routines.
-   */
-  void VerifyD2DDevice(bool aAttemptForce);
-
   void GetCommonFallbackFonts(uint32_t aCh, Script aRunScript,
-                              eFontPresentation aPresentation,
+                              FontPresentation aPresentation,
                               nsTArray<const char*>& aFontList) override;
 
-  bool CanUseHardwareVideoDecoding() override;
-
   void CompositorUpdated() override;
 
   bool DidRenderingDeviceReset(
-      DeviceResetReason* aResetReason = nullptr) override;
+      mozilla::gfx::DeviceResetReason* aResetReason = nullptr) override;
   void SchedulePaintIfDeviceReset() override;
   void CheckForContentOnlyDeviceReset();
 
@@ -193,9 +161,6 @@ class gfxWindowsPlatform final : public gfxPlatform {
 
   static bool CheckVariationFontSupport();
 
-  // Always false for content processes.
-  bool SupportsHDR() override { return mSupportsHDR; }
-
  protected:
   bool AccelerateLayersByDefault() override { return true; }
 
@@ -214,29 +179,22 @@ class gfxWindowsPlatform final : public gfxPlatform {
 
   BackendPrefsData GetBackendPrefs() const override;
 
-  void UpdateSupportsHDR();
-
-  RenderMode mRenderMode;
-  bool mSupportsHDR;
-
  private:
   void Init();
   void InitAcceleration() override;
   void InitWebRenderConfig() override;
+  void InitPlatformHardwareVideoConfig() override;
+#ifdef MOZ_WMF_CDM
+  void InitPlatformHardwarDRMConfig() override;
+#endif
 
   void InitializeDevices();
   void InitializeD3D11();
-  void InitializeD2D();
   bool InitDWriteSupport();
-  void InitGPUProcessSupport();
-
-  void DisableD2D(mozilla::gfx::FeatureStatus aStatus, const char* aMessage,
-                  const nsACString& aFailureId);
 
   void InitializeConfig();
   void InitializeD3D9Config();
   void InitializeD3D11Config();
-  void InitializeD2DConfig();
   void InitializeDirectDrawConfig();
 
   void RecordStartupTelemetry();
