@@ -9,7 +9,10 @@ use api::{PrimitiveKeyKind, FillRule, POLYGON_CLIP_VERTEX_MAX};
 use api::units::*;
 use euclid::{SideOffsets2D, Size2D};
 use malloc_size_of::MallocSizeOf;
+use crate::composite::CompositorSurfaceKind;
 use crate::clip::ClipLeafId;
+use crate::pattern::{Pattern, PatternBuilder, PatternBuilderContext, PatternBuilderState};
+use crate::quad::QuadTileClassifier;
 use crate::segment::EdgeAaSegmentMask;
 use crate::border::BorderSegmentCacheKey;
 use crate::debug_item::{DebugItem, DebugMessage};
@@ -47,6 +50,7 @@ use image::{ImageDataHandle, ImageInstance, YuvImageDataHandle};
 use line_dec::LineDecorationDataHandle;
 use picture::PictureDataHandle;
 use text_run::{TextRunDataHandle, TextRunPrimitive};
+use crate::box_shadow::BoxShadowDataHandle;
 
 pub const VECS_PER_SEGMENT: usize = 2;
 
@@ -89,6 +93,7 @@ pub struct DeferredResolve {
     pub address: GpuCacheAddress,
     pub image_properties: ImageProperties,
     pub rendering: ImageRendering,
+    pub is_composited: bool,
 }
 
 #[derive(Debug, Copy, Clone, PartialEq)]
@@ -567,6 +572,41 @@ pub struct PrimitiveTemplate {
     pub kind: PrimitiveTemplateKind,
 }
 
+impl PatternBuilder for PrimitiveTemplate {
+    fn build(
+        &self,
+        _sub_rect: Option<DeviceRect>,
+        ctx: &PatternBuilderContext,
+        _state: &mut PatternBuilderState,
+    ) -> crate::pattern::Pattern {
+        match self.kind {
+            PrimitiveTemplateKind::Clear => Pattern::clear(),
+            PrimitiveTemplateKind::Rectangle { ref color, .. } => {
+                let color = ctx.scene_properties.resolve_color(color);
+                Pattern::color(color)
+            }
+        }
+    }
+
+    fn get_base_color(
+        &self,
+        ctx: &PatternBuilderContext,
+    ) -> ColorF {
+        match self.kind {
+            PrimitiveTemplateKind::Clear => ColorF::BLACK,
+            PrimitiveTemplateKind::Rectangle { ref color, .. } => {
+                ctx.scene_properties.resolve_color(color)
+            }
+        }
+    }
+
+    fn use_shared_pattern(
+        &self,
+    ) -> bool {
+        true
+    }
+}
+
 impl ops::Deref for PrimitiveTemplate {
     type Target = PrimTemplateCommonData;
     fn deref(&self) -> &Self::Target {
@@ -635,7 +675,6 @@ impl InternablePrimitive for PrimitiveKeyKind {
         key: PrimitiveKey,
         data_handle: PrimitiveDataHandle,
         prim_store: &mut PrimitiveStore,
-        _reference_frame_relative_offset: LayoutVector2D,
     ) -> PrimitiveInstanceKind {
         match key.kind {
             PrimitiveKeyKind::Clear => {
@@ -959,7 +998,6 @@ pub enum PrimitiveInstanceKind {
         /// Handle to the common interned data for this primitive.
         data_handle: PictureDataHandle,
         pic_index: PictureIndex,
-        segment_instance_index: SegmentInstanceIndex,
     },
     /// A run of glyphs, with associated font parameters.
     TextRun {
@@ -1003,13 +1041,13 @@ pub enum PrimitiveInstanceKind {
         /// Handle to the common interned data for this primitive.
         data_handle: YuvImageDataHandle,
         segment_instance_index: SegmentInstanceIndex,
-        is_compositor_surface: bool,
+        compositor_surface_kind: CompositorSurfaceKind,
     },
     Image {
         /// Handle to the common interned data for this primitive.
         data_handle: ImageDataHandle,
         image_instance_index: ImageInstanceIndex,
-        is_compositor_surface: bool,
+        compositor_surface_kind: CompositorSurfaceKind,
     },
     /// Always rendered directly into the picture. This tends to be
     /// faster with SWGL.
@@ -1017,6 +1055,7 @@ pub enum PrimitiveInstanceKind {
         /// Handle to the common interned data for this primitive.
         data_handle: LinearGradientDataHandle,
         visible_tiles_range: GradientTileRange,
+        use_legacy_path: bool,
     },
     /// Always rendered via a cached render task. Usually faster with
     /// a GPU.
@@ -1029,11 +1068,13 @@ pub enum PrimitiveInstanceKind {
         /// Handle to the common interned data for this primitive.
         data_handle: RadialGradientDataHandle,
         visible_tiles_range: GradientTileRange,
+        use_legacy_path: bool,
     },
     ConicGradient {
         /// Handle to the common interned data for this primitive.
         data_handle: ConicGradientDataHandle,
         visible_tiles_range: GradientTileRange,
+        use_legacy_path: bool,
     },
     /// Clear out a rect, used for special effects.
     Clear {
@@ -1048,6 +1089,9 @@ pub enum PrimitiveInstanceKind {
         data_handle: BackdropRenderDataHandle,
         pic_index: PictureIndex,
     },
+    BoxShadow {
+        data_handle: BoxShadowDataHandle,
+    },
 }
 
 impl PrimitiveInstanceKind {
@@ -1148,6 +1192,9 @@ impl PrimitiveInstance {
             PrimitiveInstanceKind::BackdropRender { data_handle, .. } => {
                 data_handle.uid()
             }
+            PrimitiveInstanceKind::BoxShadow { data_handle, .. } => {
+                data_handle.uid()
+            }
         }
     }
 }
@@ -1214,8 +1261,14 @@ pub struct PrimitiveScratchBuffer {
     /// Set of sub-graphs that are required, determined during visibility pass
     pub required_sub_graphs: FastHashSet<PictureIndex>,
 
-    /// Temporary buffer for building segments in to during prepare pass
-    pub quad_segments: Vec<QuadSegment>,
+    /// Temporary buffers for building segments in to during prepare pass
+    pub quad_direct_segments: Vec<QuadSegment>,
+    pub quad_color_segments: Vec<QuadSegment>,
+    pub quad_indirect_segments: Vec<QuadSegment>,
+
+    /// A retained classifier for checking which segments of a tiled primitive
+    /// need a mask / are clipped / can be rendered directly
+    pub quad_tile_classifier: QuadTileClassifier,
 }
 
 impl Default for PrimitiveScratchBuffer {
@@ -1230,7 +1283,10 @@ impl Default for PrimitiveScratchBuffer {
             debug_items: Vec::new(),
             messages: Vec::new(),
             required_sub_graphs: FastHashSet::default(),
-            quad_segments: Vec::new(),
+            quad_direct_segments: Vec::new(),
+            quad_color_segments: Vec::new(),
+            quad_indirect_segments: Vec::new(),
+            quad_tile_classifier: QuadTileClassifier::new(),
         }
     }
 }
@@ -1244,7 +1300,9 @@ impl PrimitiveScratchBuffer {
         self.segment_instances.recycle(recycler);
         self.gradient_tiles.recycle(recycler);
         recycler.recycle_vec(&mut self.debug_items);
-        recycler.recycle_vec(&mut self.quad_segments);
+        recycler.recycle_vec(&mut self.quad_direct_segments);
+        recycler.recycle_vec(&mut self.quad_color_segments);
+        recycler.recycle_vec(&mut self.quad_indirect_segments);
     }
 
     pub fn begin_frame(&mut self) {
@@ -1253,7 +1311,9 @@ impl PrimitiveScratchBuffer {
         // location.
         self.clip_mask_instances.clear();
         self.clip_mask_instances.push(ClipMaskKind::None);
-        self.quad_segments.clear();
+        self.quad_direct_segments.clear();
+        self.quad_color_segments.clear();
+        self.quad_indirect_segments.clear();
 
         self.border_cache_handles.clear();
 
@@ -1274,7 +1334,7 @@ impl PrimitiveScratchBuffer {
         const LINE_HEIGHT: f32 = 20.0;
         const X0: f32 = 32.0;
         const Y0: f32 = 32.0;
-        let now = time::precise_time_ns();
+        let now = zeitstempel::now();
 
         let msgs_to_remove = self.messages.len().max(MSGS_TO_RETAIN) - MSGS_TO_RETAIN;
         let mut msgs_removed = 0;
@@ -1312,10 +1372,42 @@ impl PrimitiveScratchBuffer {
         }
     }
 
+    pub fn push_debug_rect_with_stroke_width(
+        &mut self,
+        rect: WorldRect,
+        border: ColorF,
+        stroke_width: f32
+    ) {
+        let top_edge = WorldRect::new(
+            WorldPoint::new(rect.min.x + stroke_width, rect.min.y),
+            WorldPoint::new(rect.max.x - stroke_width, rect.min.y + stroke_width)
+        );
+        self.push_debug_rect(top_edge * DevicePixelScale::new(1.0), 1, border, border);
+
+        let bottom_edge = WorldRect::new(
+            WorldPoint::new(rect.min.x + stroke_width, rect.max.y - stroke_width),
+            WorldPoint::new(rect.max.x - stroke_width, rect.max.y)
+        );
+        self.push_debug_rect(bottom_edge * DevicePixelScale::new(1.0), 1, border, border);
+
+        let right_edge = WorldRect::new(
+            WorldPoint::new(rect.max.x - stroke_width, rect.min.y),
+            rect.max
+        );
+        self.push_debug_rect(right_edge * DevicePixelScale::new(1.0), 1, border, border);
+
+        let left_edge = WorldRect::new(
+            rect.min,
+            WorldPoint::new(rect.min.x + stroke_width, rect.max.y)
+        );
+        self.push_debug_rect(left_edge * DevicePixelScale::new(1.0), 1, border, border);
+    }
+
     #[allow(dead_code)]
     pub fn push_debug_rect(
         &mut self,
         rect: DeviceRect,
+        thickness: i32,
         outer_color: ColorF,
         inner_color: ColorF,
     ) {
@@ -1323,6 +1415,7 @@ impl PrimitiveScratchBuffer {
             rect,
             outer_color,
             inner_color,
+            thickness,
         });
     }
 
@@ -1347,7 +1440,7 @@ impl PrimitiveScratchBuffer {
     ) {
         self.messages.push(DebugMessage {
             msg,
-            timestamp: time::precise_time_ns(),
+            timestamp: zeitstempel::now(),
         })
     }
 }
@@ -1401,6 +1494,14 @@ impl PrimitiveStore {
         }
     }
 
+    pub fn reset(&mut self) {
+        self.pictures.clear();
+        self.text_runs.clear();
+        self.images.clear();
+        self.color_bindings.clear();
+        self.linear_gradients.clear();
+    }
+
     pub fn get_stats(&self) -> PrimitiveStoreStats {
         PrimitiveStoreStats {
             picture_count: self.pictures.len(),
@@ -1419,6 +1520,12 @@ impl PrimitiveStore {
     }
 }
 
+impl Default for PrimitiveStore {
+    fn default() -> Self {
+        PrimitiveStore::new(&PrimitiveStoreStats::empty())
+    }
+}
+
 /// Trait for primitives that are directly internable.
 /// see SceneBuilder::add_primitive<P>
 pub trait InternablePrimitive: intern::Internable<InternData = ()> + Sized {
@@ -1432,7 +1539,6 @@ pub trait InternablePrimitive: intern::Internable<InternData = ()> + Sized {
         key: Self::Key,
         data_handle: intern::Handle<Self>,
         prim_store: &mut PrimitiveStore,
-        reference_frame_relative_offset: LayoutVector2D,
     ) -> PrimitiveInstanceKind;
 }
 
