@@ -9,16 +9,22 @@
 
 use crate::parser::{Parse, ParserContext};
 use crate::values::computed::basic_shape::InsetRect as ComputedInsetRect;
-use crate::values::computed::{Context, ToComputedValue};
+use crate::values::computed::{
+    Context, LengthPercentage as ComputedLengthPercentage, ToComputedValue,
+};
 use crate::values::generics::basic_shape as generic;
 use crate::values::generics::basic_shape::{Path, PolygonCoord};
+use crate::values::generics::position::GenericPositionOrAuto;
 use crate::values::generics::rect::Rect;
+use crate::values::specified::angle::Angle;
 use crate::values::specified::border::BorderRadius;
 use crate::values::specified::image::Image;
 use crate::values::specified::length::LengthPercentageOrAuto;
-use crate::values::specified::position::{Position, PositionOrAuto};
+use crate::values::specified::position::{Position, Side};
 use crate::values::specified::url::SpecifiedUrl;
+use crate::values::specified::PositionComponent;
 use crate::values::specified::{LengthPercentage, NonNegativeLengthPercentage, SVGPathData};
+use crate::values::CSSFloat;
 use crate::Zero;
 use cssparser::Parser;
 use std::fmt::{self, Write};
@@ -33,29 +39,36 @@ pub type ClipPath = generic::GenericClipPath<BasicShape, SpecifiedUrl>;
 /// A specified `shape-outside` value.
 pub type ShapeOutside = generic::GenericShapeOutside<BasicShape, Image>;
 
+/// A specified value for `at <position>` in circle() and ellipse().
+// Note: its computed value is the same as computed::position::Position. We just want to always use
+// LengthPercentage as the type of its components, for basic shapes.
+pub type RadialPosition = generic::ShapePosition<LengthPercentage>;
+
 /// A specified basic shape.
-pub type BasicShape = generic::GenericBasicShape<
-    Position,
-    LengthPercentage,
-    NonNegativeLengthPercentage,
-    BasicShapeRect,
->;
+pub type BasicShape = generic::GenericBasicShape<Angle, Position, LengthPercentage, BasicShapeRect>;
 
 /// The specified value of `inset()`.
-pub type InsetRect = generic::GenericInsetRect<LengthPercentage, NonNegativeLengthPercentage>;
+pub type InsetRect = generic::GenericInsetRect<LengthPercentage>;
 
 /// A specified circle.
-pub type Circle = generic::Circle<Position, NonNegativeLengthPercentage>;
+pub type Circle = generic::Circle<LengthPercentage>;
 
 /// A specified ellipse.
-pub type Ellipse = generic::Ellipse<Position, NonNegativeLengthPercentage>;
+pub type Ellipse = generic::Ellipse<LengthPercentage>;
 
 /// The specified value of `ShapeRadius`.
-pub type ShapeRadius = generic::ShapeRadius<NonNegativeLengthPercentage>;
+pub type ShapeRadius = generic::ShapeRadius<LengthPercentage>;
 
 /// The specified value of `Polygon`.
 pub type Polygon = generic::GenericPolygon<LengthPercentage>;
 
+/// The specified value of `PathOrShapeFunction`.
+pub type PathOrShapeFunction =
+    generic::GenericPathOrShapeFunction<Angle, Position, LengthPercentage>;
+
+/// The specified value of `ShapeCommand`.
+pub type ShapeCommand = generic::GenericShapeCommand<Angle, Position, LengthPercentage>;
+
 /// The specified value of `xywh()`.
 /// Defines a rectangle via offsets from the top and left edge of the reference box, and a
 /// specified width and height.
@@ -129,25 +142,6 @@ pub enum ShapeType {
     Outline,
 }
 
-/// The default `at <position>` if it is omitted.
-///
-/// https://github.com/w3c/csswg-drafts/issues/8695
-///
-/// FIXME: Bug 1837340. It seems we should always omit this component if the author doesn't specify
-/// it. In order to avoid changing the behavior on the shipped clip-path and shape-outside, we
-/// still use center as their default value for now.
-pub enum DefaultPosition {
-    /// Use standard default value, center, if "at <position>" is omitted.
-    Center,
-    /// The default value depends on the context. For example, offset-path:circle() may use the
-    /// value of offset-position as its default position of the circle center. So we shouldn't
-    /// assign a default value to its specified value and computed value. This makes the
-    /// serialization ignore this component (and makes this value non-interpolated with other
-    /// values which specify `at <position>`).
-    /// https://drafts.fxtf.org/motion-1/#valdef-offset-path-basic-shape
-    Context,
-}
-
 bitflags! {
     /// The flags to represent which basic shapes we would like to support.
     ///
@@ -182,25 +176,26 @@ bitflags! {
         const POLYGON = 1 << 5;
         /// path().
         const PATH = 1 << 6;
-        // TODO: Bug 1823463. Add shape().
-        // const SHAPE = 1 << 7;
+        /// shape().
+        const SHAPE = 1 << 7;
 
         /// All flags.
         const ALL =
-            Self::INSET.bits |
-            Self::XYWH.bits |
-            Self::RECT.bits |
-            Self::CIRCLE.bits |
-            Self::ELLIPSE.bits |
-            Self::POLYGON.bits |
-            Self::PATH.bits;
+            Self::INSET.bits() |
+            Self::XYWH.bits() |
+            Self::RECT.bits() |
+            Self::CIRCLE.bits() |
+            Self::ELLIPSE.bits() |
+            Self::POLYGON.bits() |
+            Self::PATH.bits() |
+            Self::SHAPE.bits();
 
         /// For shape-outside.
         const SHAPE_OUTSIDE =
-            Self::INSET.bits |
-            Self::CIRCLE.bits |
-            Self::ELLIPSE.bits |
-            Self::POLYGON.bits;
+            Self::INSET.bits() |
+            Self::CIRCLE.bits() |
+            Self::ELLIPSE.bits() |
+            Self::POLYGON.bits();
     }
 }
 
@@ -220,15 +215,7 @@ where
     loop {
         if shape.is_none() {
             shape = input
-                .try_parse(|i| {
-                    BasicShape::parse(
-                        context,
-                        i,
-                        flags,
-                        ShapeType::Filled,
-                        DefaultPosition::Center,
-                    )
-                })
+                .try_parse(|i| BasicShape::parse(context, i, flags, ShapeType::Filled))
                 .ok();
         }
 
@@ -312,7 +299,6 @@ impl BasicShape {
         input: &mut Parser<'i, 't>,
         flags: AllowedBasicShapes,
         shape_type: ShapeType,
-        default_position: DefaultPosition,
     ) -> Result<Self, ParseError<'i>> {
         let location = input.current_source_location();
         let function = input.expect_function()?.clone();
@@ -323,28 +309,22 @@ impl BasicShape {
                         .map(BasicShapeRect::Inset)
                         .map(BasicShape::Rect)
                 },
-                "xywh"
-                    if flags.contains(AllowedBasicShapes::XYWH)
-                        && static_prefs::pref!("layout.css.basic-shape-xywh.enabled") =>
-                {
+                "xywh" if flags.contains(AllowedBasicShapes::XYWH) => {
                     Xywh::parse_function_arguments(context, i)
                         .map(BasicShapeRect::Xywh)
                         .map(BasicShape::Rect)
                 },
-                "rect"
-                    if flags.contains(AllowedBasicShapes::RECT)
-                        && static_prefs::pref!("layout.css.basic-shape-rect.enabled") =>
-                {
+                "rect" if flags.contains(AllowedBasicShapes::RECT) => {
                     ShapeRectFunction::parse_function_arguments(context, i)
                         .map(BasicShapeRect::Rect)
                         .map(BasicShape::Rect)
                 },
                 "circle" if flags.contains(AllowedBasicShapes::CIRCLE) => {
-                    Circle::parse_function_arguments(context, i, default_position)
+                    Circle::parse_function_arguments(context, i)
                         .map(BasicShape::Circle)
                 },
                 "ellipse" if flags.contains(AllowedBasicShapes::ELLIPSE) => {
-                    Ellipse::parse_function_arguments(context, i, default_position)
+                    Ellipse::parse_function_arguments(context, i)
                         .map(BasicShape::Ellipse)
                 },
                 "polygon" if flags.contains(AllowedBasicShapes::POLYGON) => {
@@ -352,7 +332,17 @@ impl BasicShape {
                         .map(BasicShape::Polygon)
                 },
                 "path" if flags.contains(AllowedBasicShapes::PATH) => {
-                    Path::parse_function_arguments(i, shape_type).map(BasicShape::Path)
+                    Path::parse_function_arguments(i, shape_type)
+                        .map(PathOrShapeFunction::Path)
+                        .map(BasicShape::PathOrShape)
+                },
+                "shape"
+                    if flags.contains(AllowedBasicShapes::SHAPE)
+                        && static_prefs::pref!("layout.css.basic-shape-shape.enabled") =>
+                {
+                    generic::Shape::parse_function_arguments(context, i, shape_type)
+                        .map(PathOrShapeFunction::Shape)
+                        .map(BasicShape::PathOrShape)
                 },
                 _ => Err(location
                     .new_custom_error(StyleParseErrorKind::UnexpectedFunction(function.clone()))),
@@ -397,21 +387,61 @@ impl InsetRect {
     }
 }
 
+impl ToCss for RadialPosition {
+    fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result
+    where
+        W: Write,
+    {
+        self.horizontal.to_css(dest)?;
+        dest.write_char(' ')?;
+        self.vertical.to_css(dest)
+    }
+}
+
+fn convert_to_length_percentage<S: Side>(c: PositionComponent<S>) -> LengthPercentage {
+    use crate::values::specified::{AllowedNumericType, Percentage};
+    // Convert the value when parsing, to make sure we serialize it properly for both
+    // specified and computed values.
+    // https://drafts.csswg.org/css-shapes-1/#basic-shape-serialization
+    match c {
+        // Since <position> keywords stand in for percentages, keywords without an offset
+        // turn into percentages.
+        PositionComponent::Center => LengthPercentage::from(Percentage::new(0.5)),
+        PositionComponent::Side(keyword, None) => {
+            Percentage::new(if keyword.is_start() { 0. } else { 1. }).into()
+        },
+        // Per spec issue, https://github.com/w3c/csswg-drafts/issues/8695, the part of
+        // "avoiding calc() expressions where possible" and "avoiding calc()
+        // transformations" will be removed from the spec, and we should follow the
+        // css-values-4 for position, i.e. we make it as length-percentage always.
+        // https://drafts.csswg.org/css-shapes-1/#basic-shape-serialization.
+        // https://drafts.csswg.org/css-values-4/#typedef-position
+        PositionComponent::Side(keyword, Some(length)) => {
+            if keyword.is_start() {
+                length
+            } else {
+                length.hundred_percent_minus(AllowedNumericType::All)
+            }
+        },
+        PositionComponent::Length(length) => length,
+    }
+}
+
 fn parse_at_position<'i, 't>(
     context: &ParserContext,
     input: &mut Parser<'i, 't>,
-    default_position: DefaultPosition,
-) -> Result<PositionOrAuto, ParseError<'i>> {
+) -> Result<GenericPositionOrAuto<RadialPosition>, ParseError<'i>> {
+    use crate::values::specified::position::Position;
     if input.try_parse(|i| i.expect_ident_matching("at")).is_ok() {
-        Position::parse(context, input).map(PositionOrAuto::Position)
+        Position::parse(context, input).map(|pos| {
+            GenericPositionOrAuto::Position(RadialPosition::new(
+                convert_to_length_percentage(pos.horizontal),
+                convert_to_length_percentage(pos.vertical),
+            ))
+        })
     } else {
-        // FIXME: Bug 1837340. Per spec issue, https://github.com/w3c/csswg-drafts/issues/8695, we
-        // may not serialize the optional `at <position>` for all basic shapes. So we will drop
-        // this later.
-        match default_position {
-            DefaultPosition::Center => Ok(PositionOrAuto::Position(Position::center())),
-            DefaultPosition::Context => Ok(PositionOrAuto::Auto),
-        }
+        // `at <position>` is omitted.
+        Ok(GenericPositionOrAuto::Auto)
     }
 }
 
@@ -421,9 +451,7 @@ impl Parse for Circle {
         input: &mut Parser<'i, 't>,
     ) -> Result<Self, ParseError<'i>> {
         input.expect_function_matching("circle")?;
-        input.parse_nested_block(|i| {
-            Self::parse_function_arguments(context, i, DefaultPosition::Center)
-        })
+        input.parse_nested_block(|i| Self::parse_function_arguments(context, i))
     }
 }
 
@@ -431,12 +459,11 @@ impl Circle {
     fn parse_function_arguments<'i, 't>(
         context: &ParserContext,
         input: &mut Parser<'i, 't>,
-        default_position: DefaultPosition,
     ) -> Result<Self, ParseError<'i>> {
         let radius = input
             .try_parse(|i| ShapeRadius::parse(context, i))
             .unwrap_or_default();
-        let position = parse_at_position(context, input, default_position)?;
+        let position = parse_at_position(context, input)?;
 
         Ok(generic::Circle { radius, position })
     }
@@ -448,9 +475,7 @@ impl Parse for Ellipse {
         input: &mut Parser<'i, 't>,
     ) -> Result<Self, ParseError<'i>> {
         input.expect_function_matching("ellipse")?;
-        input.parse_nested_block(|i| {
-            Self::parse_function_arguments(context, i, DefaultPosition::Center)
-        })
+        input.parse_nested_block(|i| Self::parse_function_arguments(context, i))
     }
 }
 
@@ -458,7 +483,6 @@ impl Ellipse {
     fn parse_function_arguments<'i, 't>(
         context: &ParserContext,
         input: &mut Parser<'i, 't>,
-        default_position: DefaultPosition,
     ) -> Result<Self, ParseError<'i>> {
         let (semiaxis_x, semiaxis_y) = input
             .try_parse(|i| -> Result<_, ParseError> {
@@ -468,7 +492,7 @@ impl Ellipse {
                 ))
             })
             .unwrap_or_default();
-        let position = parse_at_position(context, input, default_position)?;
+        let position = parse_at_position(context, input)?;
 
         Ok(generic::Ellipse {
             semiaxis_x,
@@ -478,7 +502,11 @@ impl Ellipse {
     }
 }
 
-fn parse_fill_rule<'i, 't>(input: &mut Parser<'i, 't>, shape_type: ShapeType) -> FillRule {
+fn parse_fill_rule<'i, 't>(
+    input: &mut Parser<'i, 't>,
+    shape_type: ShapeType,
+    expect_comma: bool,
+) -> FillRule {
     match shape_type {
         // Per [1] and [2], we ignore `<fill-rule>` for outline shapes, so always use a default
         // value.
@@ -496,7 +524,9 @@ fn parse_fill_rule<'i, 't>(input: &mut Parser<'i, 't>, shape_type: ShapeType) ->
         ShapeType::Filled => input
             .try_parse(|i| -> Result<_, ParseError> {
                 let fill = FillRule::parse(i)?;
-                i.expect_comma()?;
+                if expect_comma {
+                    i.expect_comma()?;
+                }
                 Ok(fill)
             })
             .unwrap_or_default(),
@@ -520,7 +550,7 @@ impl Polygon {
         input: &mut Parser<'i, 't>,
         shape_type: ShapeType,
     ) -> Result<Self, ParseError<'i>> {
-        let fill = parse_fill_rule(input, shape_type);
+        let fill = parse_fill_rule(input, shape_type, true /* has comma */);
         let coordinates = input
             .parse_comma_separated(|i| {
                 Ok(PolygonCoord(
@@ -542,7 +572,7 @@ impl Path {
     ) -> Result<Self, ParseError<'i>> {
         use crate::values::specified::svg_path::AllowEmpty;
 
-        let fill = parse_fill_rule(input, shape_type);
+        let fill = parse_fill_rule(input, shape_type, true /* has comma */);
         let path = SVGPathData::parse(input, AllowEmpty::No)?;
         Ok(Path { fill, path })
     }
@@ -705,3 +735,332 @@ impl ToComputedValue for BasicShapeRect {
         Self::Inset(ToComputedValue::from_computed_value(computed))
     }
 }
+
+impl generic::Shape<Angle, Position, LengthPercentage> {
+    /// Parse the inner arguments of a `shape` function.
+    /// shape() = shape(<fill-rule>? from <coordinate-pair>, <shape-command>#)
+    fn parse_function_arguments<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+        shape_type: ShapeType,
+    ) -> Result<Self, ParseError<'i>> {
+        let fill = parse_fill_rule(input, shape_type, false /* no following comma */);
+
+        let mut first = true;
+        let commands = input.parse_comma_separated(|i| {
+            if first {
+                first = false;
+
+                // The starting point for the first shape-command. It adds an initial absolute
+                // moveto to the list of path data commands, with the <coordinate-pair> measured
+                // from the top-left corner of the reference
+                i.expect_ident_matching("from")?;
+                Ok(ShapeCommand::Move {
+                    point: generic::CommandEndPoint::parse_endpoint_as_abs(context, i)?,
+                })
+            } else {
+                // The further path data commands.
+                ShapeCommand::parse(context, i)
+            }
+        })?;
+
+        // We must have one starting point and at least one following <shape-command>.
+        if commands.len() < 2 {
+            return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+        }
+
+        Ok(Self {
+            fill,
+            commands: commands.into(),
+        })
+    }
+}
+
+impl Parse for ShapeCommand {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        use crate::values::generics::basic_shape::{
+            ArcRadii, ArcSize, ArcSweep, AxisEndPoint, CommandEndPoint, ControlPoint,
+        };
+
+        // <shape-command> = <move-command> | <line-command> | <hv-line-command> |
+        //                   <curve-command> | <smooth-command> | <arc-command> | close
+        Ok(try_match_ident_ignore_ascii_case! { input,
+            "close" => Self::Close,
+            "move" => {
+                let point = CommandEndPoint::parse(context, input)?;
+                Self::Move { point }
+            },
+            "line" => {
+                let point = CommandEndPoint::parse(context, input)?;
+                Self::Line { point }
+            },
+            "hline" => {
+                let x = AxisEndPoint::parse_hline(context, input)?;
+                Self::HLine { x }
+            },
+            "vline" => {
+                let y = AxisEndPoint::parse_vline(context, input)?;
+                Self::VLine { y }
+            },
+            "curve" => {
+                let point = CommandEndPoint::parse(context, input)?;
+                input.expect_ident_matching("with")?;
+                let control1 = ControlPoint::parse(context, input, point.is_abs())?;
+                if input.try_parse(|i| i.expect_delim('/')).is_ok() {
+                    let control2 = ControlPoint::parse(context, input, point.is_abs())?;
+                    Self::CubicCurve {
+                        point,
+                        control1,
+                        control2,
+                    }
+                } else {
+                    Self::QuadCurve {
+                        point,
+                        control1,
+                    }
+                }
+            },
+            "smooth" => {
+                let point = CommandEndPoint::parse(context, input)?;
+                if input.try_parse(|i| i.expect_ident_matching("with")).is_ok() {
+                    let control2 = ControlPoint::parse(context, input, point.is_abs())?;
+                    Self::SmoothCubic {
+                        point,
+                        control2,
+                    }
+                } else {
+                    Self::SmoothQuad { point }
+                }
+            },
+            "arc" => {
+                let point = CommandEndPoint::parse(context, input)?;
+                input.expect_ident_matching("of")?;
+                let rx = LengthPercentage::parse(context, input)?;
+                let ry = input.try_parse(|i| LengthPercentage::parse(context, i)).ok();
+                let radii = ArcRadii { rx, ry: ry.into() };
+
+                // [<arc-sweep> || <arc-size> || rotate <angle>]?
+                let mut arc_sweep = None;
+                let mut arc_size = None;
+                let mut rotate = None;
+                loop {
+                    if arc_sweep.is_none() {
+                        arc_sweep = input.try_parse(ArcSweep::parse).ok();
+                    }
+
+                    if arc_size.is_none() {
+                        arc_size = input.try_parse(ArcSize::parse).ok();
+                        if arc_size.is_some() {
+                            continue;
+                        }
+                    }
+
+                    if rotate.is_none()
+                        && input
+                            .try_parse(|i| i.expect_ident_matching("rotate"))
+                            .is_ok()
+                    {
+                        rotate = Some(Angle::parse(context, input)?);
+                        continue;
+                    }
+                    break;
+                }
+                Self::Arc {
+                    point,
+                    radii,
+                    arc_sweep: arc_sweep.unwrap_or(ArcSweep::Ccw),
+                    arc_size: arc_size.unwrap_or(ArcSize::Small),
+                    rotate: rotate.unwrap_or(Angle::zero()),
+                }
+            },
+        })
+    }
+}
+
+impl Parse for generic::CoordinatePair<LengthPercentage> {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        let x = LengthPercentage::parse(context, input)?;
+        let y = LengthPercentage::parse(context, input)?;
+        Ok(Self::new(x, y))
+    }
+}
+
+impl generic::ControlPoint<Position, LengthPercentage> {
+    /// Parse <control-point> = [ <position> | <relative-control-point> ]
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+        is_end_point_abs: bool,
+    ) -> Result<Self, ParseError<'i>> {
+        use generic::ControlReference;
+        let coord = input.try_parse(|i| generic::CoordinatePair::parse(context, i));
+
+        // Parse <position>
+        if is_end_point_abs && coord.is_err() {
+            let pos = Position::parse(context, input)?;
+            return Ok(Self::Absolute(pos));
+        }
+
+        // Parse <relative-control-point> = <coordinate-pair> [from [ start | end | origin ]]?
+        let coord = coord?;
+        let mut reference = if is_end_point_abs {
+            ControlReference::Origin
+        } else {
+            ControlReference::Start
+        };
+        if input.try_parse(|i| i.expect_ident_matching("from")).is_ok() {
+            reference = ControlReference::parse(input)?;
+        }
+
+        Ok(Self::Relative(generic::RelativeControlPoint {
+            coord,
+            reference,
+        }))
+    }
+}
+
+impl Parse for generic::CommandEndPoint<Position, LengthPercentage> {
+    /// Parse <command-end-point> = to <position> | by <coordinate-pair>
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        if ByTo::parse(input)?.is_abs() {
+            Self::parse_endpoint_as_abs(context, input)
+        } else {
+            let point = generic::CoordinatePair::parse(context, input)?;
+            Ok(Self::ByCoordinate(point))
+        }
+    }
+}
+
+impl generic::CommandEndPoint<Position, LengthPercentage> {
+    /// Parse <command-end-point> = to <position>
+    fn parse_endpoint_as_abs<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        let point = Position::parse(context, input)?;
+        Ok(generic::CommandEndPoint::ToPosition(point))
+    }
+}
+
+impl generic::AxisEndPoint<LengthPercentage> {
+    /// Parse <horizontal-line-command>
+    pub fn parse_hline<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        use cssparser::Token;
+        use generic::{AxisPosition, AxisPositionKeyword};
+
+        // If the command is relative, parse for <length-percentage> only.
+        if !ByTo::parse(input)?.is_abs() {
+            return Ok(Self::ByCoordinate(LengthPercentage::parse(context, input)?));
+        }
+
+        let x = AxisPosition::parse(context, input)?;
+        if let AxisPosition::Keyword(
+            _word @ (AxisPositionKeyword::Top
+            | AxisPositionKeyword::Bottom
+            | AxisPositionKeyword::YStart
+            | AxisPositionKeyword::YEnd),
+        ) = &x
+        {
+            let location = input.current_source_location();
+            let token = Token::Ident(x.to_css_string().into());
+            return Err(location.new_unexpected_token_error(token));
+        }
+        Ok(Self::ToPosition(x))
+    }
+
+    /// Parse <vertical-line-command>
+    pub fn parse_vline<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        use cssparser::Token;
+        use generic::{AxisPosition, AxisPositionKeyword};
+
+        // If the command is relative, parse for <length-percentage> only.
+        if !ByTo::parse(input)?.is_abs() {
+            return Ok(Self::ByCoordinate(LengthPercentage::parse(context, input)?));
+        }
+
+        let y = AxisPosition::parse(context, input)?;
+        if let AxisPosition::Keyword(
+            _word @ (AxisPositionKeyword::Left
+            | AxisPositionKeyword::Right
+            | AxisPositionKeyword::XStart
+            | AxisPositionKeyword::XEnd),
+        ) = &y
+        {
+            // Return an error if we parsed a different keyword.
+            let location = input.current_source_location();
+            let token = Token::Ident(y.to_css_string().into());
+            return Err(location.new_unexpected_token_error(token));
+        }
+        Ok(Self::ToPosition(y))
+    }
+}
+
+impl ToComputedValue for generic::AxisPosition<LengthPercentage> {
+    type ComputedValue = generic::AxisPosition<ComputedLengthPercentage>;
+
+    fn to_computed_value(&self, context: &Context) -> Self::ComputedValue {
+        match self {
+            Self::LengthPercent(lp) => {
+                Self::ComputedValue::LengthPercent(lp.to_computed_value(context))
+            },
+            Self::Keyword(word) => {
+                let lp = LengthPercentage::Percentage(word.as_percentage());
+                Self::ComputedValue::LengthPercent(lp.to_computed_value(context))
+            },
+        }
+    }
+
+    fn from_computed_value(computed: &Self::ComputedValue) -> Self {
+        match computed {
+            Self::ComputedValue::LengthPercent(lp) => {
+                Self::LengthPercent(LengthPercentage::from_computed_value(lp))
+            },
+            _ => unreachable!("Invalid state: computed value cannot be a keyword."),
+        }
+    }
+}
+
+impl ToComputedValue for generic::AxisPosition<CSSFloat> {
+    type ComputedValue = Self;
+
+    fn to_computed_value(&self, _context: &Context) -> Self {
+        *self
+    }
+
+    fn from_computed_value(computed: &Self) -> Self {
+        *computed
+    }
+}
+
+/// This determines whether the command is absolutely or relatively positioned.
+/// https://drafts.csswg.org/css-shapes-1/#typedef-shape-command-end-point
+#[derive(Clone, Copy, Debug, Parse, PartialEq)]
+enum ByTo {
+    /// Command is relative to the command’s starting point.
+    By,
+    /// Command is relative to the top-left corner of the reference box.
+    To,
+}
+
+impl ByTo {
+    /// Return true if it is absolute, i.e. it is To.
+    #[inline]
+    pub fn is_abs(&self) -> bool {
+        matches!(self, ByTo::To)
+    }
+}
