@@ -189,58 +189,31 @@
       ")")
 
 // Note: this should be kept up-to-date with the definition in
-//       nsPlacesAutoComplete.js.
-#define CREATE_MOZ_OPENPAGES_TEMP              \
-  nsLiteralCString(                            \
-      "CREATE TEMP TABLE moz_openpages_temp (" \
-      "  url TEXT"                             \
-      ", userContextId INTEGER"                \
-      ", open_count INTEGER"                   \
-      ", PRIMARY KEY (url, userContextId)"     \
+//       PlacesUtils.sys.mjs.
+#define CREATE_MOZ_OPENPAGES_TEMP                   \
+  nsLiteralCString(                                 \
+      "CREATE TEMP TABLE moz_openpages_temp ("      \
+      "  url TEXT NOT NULL"                         \
+      ", userContextId INTEGER NOT NULL"            \
+      ", groupId TEXT NOT NULL"                     \
+      ", open_count INTEGER"                        \
+      ", PRIMARY KEY (url, userContextId, groupId)" \
       ")")
 
-// This table is used, along with moz_places_afterinsert_trigger, to update
-// origins after places removals. During an INSERT into moz_places, origins are
-// accumulated in this table, then a DELETE FROM moz_updateoriginsinsert_temp
-// will take care of updating the moz_origins table for every new origin. See
-// CREATE_PLACES_AFTERINSERT_TRIGGER in nsPlacestriggers.h for details.
-#define CREATE_UPDATEORIGINSINSERT_TEMP                   \
-  nsLiteralCString(                                       \
-      "CREATE TEMP TABLE moz_updateoriginsinsert_temp ( " \
-      "place_id INTEGER PRIMARY KEY, "                    \
-      "prefix TEXT NOT NULL, "                            \
-      "host TEXT NOT NULL, "                              \
-      "frecency INTEGER NOT NULL "                        \
-      ") ")
-
-// This table is used in a similar way to moz_updateoriginsinsert_temp, but for
-// deletes, and triggered via moz_places_afterdelete_trigger.
-//
-// When rows are added to this table, moz_places.origin_id may be null.  That's
-// why this table uses prefix + host as its primary key, not origin_id.
+// This table is used to remove orphan origins after pages are removed from
+// moz_places. Insertions are made by moz_places_afterdelete_trigger.
+// This allows for more performant handling of batch removals, since we'll look
+// for orphan origins only once, instead of doing it for each page removal.
+// The downside of this approach is that after the removal is complete the
+// consumer must remember to also delete from this table, and a trigger will
+// take care of orphans.
 #define CREATE_UPDATEORIGINSDELETE_TEMP                   \
   nsLiteralCString(                                       \
       "CREATE TEMP TABLE moz_updateoriginsdelete_temp ( " \
-      "prefix TEXT NOT NULL, "                            \
-      "host TEXT NOT NULL, "                              \
-      "frecency_delta INTEGER NOT NULL, "                 \
-      "PRIMARY KEY (prefix, host) "                       \
-      ") WITHOUT ROWID ")
-
-// This table is used in a similar way to moz_updateoriginsinsert_temp, but for
-// updates to places' frecencies, and triggered via
-// moz_places_afterupdate_frecency_trigger.
-//
-// When rows are added to this table, moz_places.origin_id may be null.  That's
-// why this table uses prefix + host as its primary key, not origin_id.
-#define CREATE_UPDATEORIGINSUPDATE_TEMP                   \
-  nsLiteralCString(                                       \
-      "CREATE TEMP TABLE moz_updateoriginsupdate_temp ( " \
-      "prefix TEXT NOT NULL, "                            \
-      "host TEXT NOT NULL, "                              \
-      "frecency_delta INTEGER NOT NULL, "                 \
-      "PRIMARY KEY (prefix, host) "                       \
-      ") WITHOUT ROWID ")
+      "  prefix TEXT NOT NULL, "                          \
+      "  host TEXT NOT NULL, "                            \
+      "  PRIMARY KEY (prefix, host) "                     \
+      ") WITHOUT ROWID")
 
 // This table would not be strictly needed for functionality since it's just
 // mimicking moz_places, though it's great for database portability.
@@ -271,6 +244,7 @@
       "root INTEGER NOT NULL DEFAULT 0, "      \
       "color INTEGER, "                        \
       "expire_ms INTEGER NOT NULL DEFAULT 0, " \
+      "flags INTEGER NOT NULL DEFAULT 0, "     \
       "data BLOB "                             \
       ") ")
 
@@ -298,12 +272,6 @@
       "value NOT NULL"          \
       ") WITHOUT ROWID ")
 
-// Keys in the moz_meta table.
-#define MOZ_META_KEY_ORIGIN_FRECENCY_COUNT "origin_frecency_count"
-#define MOZ_META_KEY_ORIGIN_FRECENCY_SUM "origin_frecency_sum"
-#define MOZ_META_KEY_ORIGIN_FRECENCY_SUM_OF_SQUARES \
-  "origin_frecency_sum_of_squares"
-
 // This table holds history interactions that will be used to achieve improved
 // history recalls.
 #define CREATE_MOZ_PLACES_METADATA                                           \
@@ -342,4 +310,38 @@
       "  hash TEXT PRIMARY KEY "                              \
       ") WITHOUT ROWID")
 
+#define CREATE_MOZ_NEWTAB_STORY_CLICK            \
+  nsLiteralCString(                              \
+      "CREATE TABLE moz_newtab_story_click ( "   \
+      "  feature TEXT NOT NULL, "                \
+      "  timestamp_s INTEGER NOT NULL, "         \
+      "  card_format_enum INTEGER NOT NULL, "    \
+      "  position INTEGER NOT NULL, "            \
+      "  section_position INTEGER NOT NULL, "    \
+      "  feature_value REAL NOT NULL DEFAULT 1 " \
+      ")")
+
+#define CREATE_MOZ_NEWTAB_STORY_IMPRESSION          \
+  nsLiteralCString(                                 \
+      "CREATE TABLE moz_newtab_story_impression ( " \
+      "  feature TEXT NOT NULL, "                   \
+      "  timestamp_s INTEGER NOT NULL, "            \
+      "  card_format_enum INTEGER NOT NULL, "       \
+      "  position INTEGER NOT NULL, "               \
+      "  section_position INTEGER NOT NULL, "       \
+      "  feature_value REAL NOT NULL DEFAULT 1 "    \
+      ")")
+
+#define CREATE_MOZ_NEWTAB_SHORTCUTS_INTERACTION                          \
+  nsLiteralCString(                                                      \
+      "CREATE TABLE moz_newtab_shortcuts_interaction ( "                 \
+      "  id INTEGER PRIMARY KEY, "                                       \
+      "  place_id INTEGER NOT NULL REFERENCES moz_places(id) ON DELETE " \
+      "CASCADE, "                                                        \
+      "  event_type INTEGER NOT NULL, "                                  \
+      "  timestamp_s INTEGER NOT NULL, "                                 \
+      "  pinned INTEGER NOT NULL CHECK (pinned IN (0, 1)), "             \
+      "  tile_position INTEGER NOT NULL"                                 \
+      ")")
+
 #endif  // __nsPlacesTables_h__
